import logging

from django.apps import apps
from django.contrib.auth import get_user_model

from mayan.celery import app

from .classes import DocumentVersionExporter

logger = logging.getLogger(name=__name__)


@app.task(ignore_result=True)
def task_document_version_export(
    document_version_id, organization_installation_url=None, user_id=None
):
    DocumentVersion = apps.get_model(
        app_label='documents', model_name='DocumentVersion'
    )
    User = get_user_model()

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    document_version = DocumentVersion.objects.get(
        pk=document_version_id
    )

    document_version_exporter = DocumentVersionExporter(document_version=document_version)
    document_version_exporter.export_to_download_file(
        organization_installation_url=organization_installation_url,
        user=user
    )



from django.db import migrations, models
import django.db.models.deletion

import mayan.apps.databases.model_mixins


class Migration(migrations.Migration):
    dependencies = [
        ('document_states', '0027_workflowinstance_datetime')
    ]

    operations = [
        migrations.CreateModel(
            name='WorkflowStateEscalation',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'condition', models.TextField(
                        blank=True, help_text="The condition that will "
                        "determine if this object is executed or not. "
                        "Conditions that do not return any value, that "
                        "return the Python logical None, or an empty "
                        "string ('') are considered to be logical false, "
                        "any other value is considered to be the logical "
                        "true.", verbose_name='Condition'
                    )
                ),
                (
                    'priority', models.IntegerField(
                        blank=True, db_index=True, default=0,
                        help_text='Determine the order in which the '
                        'escalations will be evaluated. Escalations with '
                        'a lower priority number are executed before '
                        'escalations with a higher priority number.',
                        verbose_name='Priority'
                    )
                ),
                (
                    'enabled', models.BooleanField(
                        default=True, help_text='Enable automatic '
                        'transition the workflow after a specified '
                        'amount of time has elapsed in the state without '
                        'change.', verbose_name='Enabled'
                    )
                ),
                (
                    'unit', models.CharField(
                        blank=True, choices=[
                            ('days', 'Days'), ('hours', 'Hours'),
                            ('minutes', 'Minutes')
                        ], default='days', max_length=32,
                        verbose_name='Time unit'
                    )
                ),
                (
                    'amount', models.PositiveIntegerField(
                        blank=True, default=1, help_text='Amount of the '
                        'selected escalation units of time.',
                        verbose_name='Amount'
                    )
                ),
                (
                    'comment', models.TextField(
                        blank=True, help_text='Comment to save to the '
                        'workflow instance when the escalation is executed.',
                        verbose_name='Comment'
                    )
                ),
                (
                    'state', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='escalations',
                        to='document_states.workflowstate',
                        verbose_name='Workflow state'
                    )
                ),
                (
                    'transition', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='escalations',
                        to='document_states.workflowtransition',
                        verbose_name='Transition'
                    )
                )
            ],
            options={
                'verbose_name': 'Workflow state escalation',
                'verbose_name_plural': 'Workflow state escalations',
                'ordering': ('priority',),
                'unique_together': {
                    ('state', 'transition')
                }
            },
            bases=(
                mayan.apps.databases.model_mixins.ExtraDataModelMixin,
                models.Model
            )
        )
    ]



from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import AuthenticationForm
from django.utils.translation import gettext_lazy as _

from mayan.apps.user_management.querysets import get_user_queryset
from mayan.apps.views.forms import FilteredSelectionForm

from .permissions import permission_users_impersonate


class AuthenticationFormBase(forms.Form):
    _label = None
    PASSWORD_FIELD = 'username'

    def __init__(
        self, data, files, prefix, initial, request=None, wizard=None
    ):
        self.request = request
        self.user_cache = None
        self.wizard = wizard

        super().__init__(
            data=data, files=files, prefix=prefix, initial=initial
        )

    def get_user(self):
        return self.user_cache


class AuthenticationFormMixinRememberMe(forms.Form):
    _form_field_name_remember_me = 'remember_me'
    remember_me = forms.BooleanField(
        label=_(message='Remember me'), required=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        field_order = [
            field for field in self.fields if field != self._form_field_name_remember_me
        ]
        field_order.append(self._form_field_name_remember_me)

        self.order_fields(field_order=field_order)


class AuthenticationFormEmailPassword(
    AuthenticationFormMixinRememberMe, AuthenticationForm
):
    """
    A form to use email address authentication.
    """
    PASSWORD_FIELD = 'email'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        UserModel = get_user_model()

        self.username_field = UserModel._meta.get_field(field_name='email')
        username_max_length = self.username_field.max_length or 254
        self.fields['username'].max_length = username_max_length
        self.fields['username'].widget.attrs['maxlength'] = username_max_length
        self.fields['username'].label = self.username_field.verbose_name


class AuthenticationFormUsernamePassword(
    AuthenticationFormMixinRememberMe, AuthenticationForm
):
    """
    Modified authentication form to include the "Remember me" field.
    """
    PASSWORD_FIELD = 'username'


class UserImpersonationOptionsForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['permanent'] = forms.BooleanField(
            label=_(message='Permanent'), help_text=_(
                'If selected, disables ending impersonation.'
            ), required=False
        )


class UserImpersonationSelectionForm(
    FilteredSelectionForm, UserImpersonationOptionsForm
):
    class Meta:
        allow_multiple = False
        field_name = 'user_to_impersonate'
        label = _(message='User')
        queryset = get_user_queryset().none()
        permission = permission_users_impersonate
        required = True
        widget_attributes = {'class': 'select2'}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        queryset = get_user_queryset().exclude(
            pk=kwargs['user'].pk
        )
        self.fields['user_to_impersonate'].queryset = queryset
        self.order_fields(
            field_order=('user_to_impersonate', 'permanent')
        )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('metadata', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='metadatatype',
            name='validation',
            field=models.CharField(
                blank=True, max_length=64,
                verbose_name='Validation function name',
                choices=[
                    (
                        b'metadata.validators.DateAndTimeValidator',
                        b'metadata.validators.DateAndTimeValidator'
                    ),
                    (
                        b'metadata.validators.DateValidator',
                        b'metadata.validators.DateValidator'
                    ),
                    (
                        b'metadata.validators.TimeValidator',
                        b'metadata.validators.TimeValidator'
                    )
                ]
            ),
            preserve_default=True,
        ),
    ]



from django.utils.translation import gettext_lazy as _

SOURCE_UNCOMPRESS_CHOICE_ALWAYS = 'y'
SOURCE_UNCOMPRESS_CHOICE_NEVER = 'n'
SOURCE_UNCOMPRESS_CHOICE_ASK = 'a'

SOURCE_UNCOMPRESS_NON_INTERACTIVE_CHOICES = (
    (SOURCE_UNCOMPRESS_CHOICE_ALWAYS, _(message='Always')),
    (SOURCE_UNCOMPRESS_CHOICE_NEVER, _(message='Never'))
)

SOURCE_UNCOMPRESS_INTERACTIVE_CHOICES = (
    (SOURCE_UNCOMPRESS_CHOICE_ALWAYS, _(message='Always')),
    (SOURCE_UNCOMPRESS_CHOICE_NEVER, _(message='Never')),
    (SOURCE_UNCOMPRESS_CHOICE_ASK, _(message='Ask user'))
)



from django.core.files import File

from rest_framework import status
from rest_framework.response import Response

from mayan.apps.sources.source_backend_actions.interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestViewForm,
    SourceBackendActionInterfaceTask
)
from mayan.apps.storage.models import SharedUploadedFile

from .arguments import (
    argument_file, argument_file_object, argument_shared_uploaded_file_id
)


class SourceBackendActionMixinFileUser:
    """
    Process a user supplied file object for upload.
    """
    accept_files = True

    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                file_object = argument_file_object

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_object'] = File(
                    file=self.context['file_object']
                )

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                file = argument_file

            def process_action_data(self):
                super().process_action_data()

                self.interface_result = Response(
                    status=status.HTTP_202_ACCEPTED
                )

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['file_object'] = self.context['file']

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                shared_uploaded_file_id = argument_shared_uploaded_file_id

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['shared_uploaded_file_id'] = self.context['shared_uploaded_file_id']

        class View(SourceBackendActionInterfaceRequestViewForm):
            def process_interface_context(self):
                super().process_interface_context()

                form_file = self.context['forms']['source_form'].cleaned_data['file']

                self.action_kwargs['file_object'] = form_file

    def _background_task(self, shared_uploaded_file_id, **kwargs):
        result = super()._background_task(**kwargs)

        result['server_upload_entry_list'] = (
            {'shared_uploaded_file_id': shared_uploaded_file_id},
        )

        return result

    def _execute(self, file_object, **kwargs):
        self.shared_uploaded_file = SharedUploadedFile.objects.create(
            file=file_object
        )

        super()._execute(**kwargs)

    def get_task_kwargs(self, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['shared_uploaded_file_id'] = self.shared_uploaded_file.pk

        return result



from django.utils.translation import gettext_lazy as _

from .classes import DocumentVersionModification
from .tasks import (
    task_document_version_page_list_append,
    task_document_version_page_list_reset
)


class DocumentVersionModificationPagesAppend(DocumentVersionModification):
    label = _(message='Append all file pages')
    description = _(
        'The current pages will be deleted and then all the '
        'document file pages will be appended as pages of this '
        'document version.'
    )

    @staticmethod
    def execute(document_version, user):
        task_document_version_page_list_append.apply_async(
            kwargs={
                'document_version_id': document_version.pk,
                'user_id': user.pk
            }
        )


class DocumentVersionModificationPagesReset(DocumentVersionModification):
    label = _(message='Reset pages to latest file')
    description = _(message='Match all pages to that of the latest document file.')

    @staticmethod
    def execute(document_version, user):
        task_document_version_page_list_reset.apply_async(
            kwargs={
                'document_version_id': document_version.pk,
                'user_id': user.pk
            }
        )



from django.utils.translation import gettext_lazy as _

from .classes import MailerBackendBaseEmail, MailerBackendCredentials


class DjangoSMTP(MailerBackendCredentials):
    """
    Backend that wraps Django's SMTP backend.
    """
    class_path = 'django.core.mail.backends.smtp.EmailBackend'
    label = _(message='Django SMTP backend')

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()

        fields.update(
            {
                'host': {
                    'label': _(message='Host'),
                    'class': 'django.forms.CharField',
                    'default': 'localhost',
                    'help_text': _(
                        message='The host to use for sending email.'
                    ),
                    'kwargs': {
                        'max_length': 48
                    }, 'required': True
                }, 'port': {
                    'label': _(message='Port'),
                    'class': 'django.forms.IntegerField', 'default': 25,
                    'help_text': _(
                        message='Port to use for the SMTP server.'
                    ),
                    'required': True
                }, 'use_tls': {
                    'label': _(message='Use TLS'),
                    'class': 'django.forms.BooleanField', 'default': False,
                    'help_text': _(
                        message='Whether to use a TLS (secure) '
                        'connection when talking to the SMTP server. This '
                        'is used for explicit TLS connections, generally '
                        'on port 587.'
                    ), 'required': False
                }, 'use_ssl': {
                    'label': _(message='Use SSL'),
                    'class': 'django.forms.BooleanField', 'default': False,
                    'help_text': _(
                        message='Whether to use an implicit TLS (secure) '
                        'connection when talking to the SMTP server. In '
                        'most email documentation this type of TLS '
                        'connection is referred to as SSL. It is generally '
                        'used on port 465. If you are experiencing '
                        'problems, see the explicit TLS setting '
                        '"Use TLS". Note that "Use TLS" and "Use SSL" '
                        'are mutually exclusive, so only set one of those '
                        'settings to True.'
                    ), 'required': False
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _('Server'), {
                    'fields': ('host', 'port', 'use_ssl', 'use_tls')
                }
            ),
        )

        return fieldsets

    def get_connection_kwargs(self):
        result = super().get_connection_kwargs()

        result.update(
            {
                'host': self.kwargs['host'],
                'port': self.kwargs['port'],
                'use_ssl': self.kwargs.get('use_ssl'),
                'use_tls': self.kwargs.get('use_tls')
            }
        )

        return result


class DjangoFileBased(MailerBackendBaseEmail):
    """
    Mailing backend that wraps Django's file based email backend.
    """
    class_path = 'django.core.mail.backends.filebased.EmailBackend'
    label = _(message='Django file based backend')

    @classmethod
    def get_form_fields(cls):
        fields = super().get_form_fields()
        fields.update(
            {
                'file_path': {
                    'label': _(message='File path'),
                    'class': 'django.forms.CharField', 'kwargs': {
                        'max_length': 48
                    }
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Storage'), {
                    'fields': ('file_path',)
                }
            ),
        )

        return fieldsets

    def get_connection_kwargs(self):
        result = super().get_connection_kwargs()

        result.update(
            {
                'file_path': self.kwargs['file_path']
            }
        )

        return result



from django.apps import apps
from django.contrib.auth import get_user_model
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.dashboards.classes import DashboardWidgetNumeric

from .icons import icon_group_list, icon_user_list
from .permissions import permission_group_view, permission_user_view


class DashboardWidgetUserTotal(DashboardWidgetNumeric):
    icon = icon_user_list
    label = _(message='Total users')
    link = reverse_lazy(viewname='user_management:user_list')

    def get_count(self):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        return AccessControlList.objects.restrict_queryset(
            permission=permission_user_view,
            queryset=get_user_model().objects.all(), user=self.request.user
        ).count()


class DashboardWidgetGroupTotal(DashboardWidgetNumeric):
    icon = icon_group_list
    label = _(message='Total groups')
    link = reverse_lazy(viewname='user_management:group_list')

    def get_count(self):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        Group = apps.get_model(
            app_label='auth', model_name='Group'
        )
        return AccessControlList.objects.restrict_queryset(
            permission=permission_group_view, queryset=Group.objects.all(),
            user=self.request.user
        ).count()



import os
from pathlib import Path

from django.conf import settings
from django.template import loader
from django.template.base import Template
from django.template.context import Context
from django.urls import include, re_path
from django.utils.html import mark_safe
from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.classes import BaseBackend
from mayan.apps.common.menus import menu_tools
from mayan.apps.common.serialization import yaml_dump, yaml_load
from mayan.apps.task_manager.classes import Worker
from mayan.settings.literals import (
    DEFAULT_DATABASE_NAME, DEFAULT_DATABASE_PASSWORD, DEFAULT_DATABASE_USER,
    DEFAULT_DIRECTORY_INSTALLATION, DEFAULT_ELASTICSEARCH_PASSWORD,
    DEFAULT_KEYCLOAK_ADMIN, DEFAULT_KEYCLOAK_ADMIN_PASSWORD,
    DEFAULT_KEYCLOAK_DATABASE_HOST, DEFAULT_KEYCLOAK_DATABASE_NAME,
    DEFAULT_KEYCLOAK_DATABASE_PASSWORD, DEFAULT_KEYCLOAK_DATABASE_USERNAME,
    DEFAULT_OS_USERNAME, DEFAULT_RABBITMQ_PASSWORD, DEFAULT_RABBITMQ_USER,
    DEFAULT_RABBITMQ_VHOST, DEFAULT_REDIS_PASSWORD,
    DEFAULT_USER_SETTINGS_FOLDER, DOCKER_DIND_IMAGE_VERSION,
    DOCKER_ELASTIC_IMAGE_NAME, DOCKER_ELASTIC_IMAGE_TAG,
    DOCKER_IMAGE_MAYAN_NAME, DOCKER_IMAGE_MAYAN_TAG,
    DOCKER_KEYCLOAK_IMAGE_NAME, DOCKER_KEYCLOAK_IMAGE_TAG,
    DOCKER_KEYCLOAK_POSTGRES_IMAGE_NAME, DOCKER_KEYCLOAK_POSTGRES_IMAGE_TAG,
    DOCKER_LINUX_IMAGE_VERSION, DOCKER_MYSQL_IMAGE_VERSION,
    DOCKER_POSTGRESQL_IMAGE_NAME, DOCKER_POSTGRESQL_IMAGE_TAG,
    DOCKER_RABBITMQ_IMAGE_NAME, DOCKER_RABBITMQ_IMAGE_TAG,
    DOCKER_REDIS_IMAGE_NAME, DOCKER_REDIS_IMAGE_TAG,
    DOCKER_TRAEFIK_IMAGE_NAME, DOCKER_TRAEFIK_IMAGE_TAG,
    GITLAB_CI_BRANCH_BUILDS_DOCKER, GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION,
    GITLAB_CI_BRANCH_BUILDS_PYTHON, GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO,
    GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING,
    GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR,
    GITLAB_CI_BRANCH_RELEASES_ALL_MINOR,
    GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR,
    GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR,
    GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION,
    GITLAB_CI_BRANCH_RELEASES_NIGHTLY,
    GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR,
    GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR,
    GITLAB_CI_BRANCH_RELEASES_STAGING, GITLAB_CI_BRANCH_RELEASES_TESTING,
    GITLAB_CI_BRANCH_TESTS_ALL, GITLAB_CI_BRANCH_TESTS_DOCKER,
    GITLAB_CI_BRANCH_TESTS_PYTHON_ALL, GITLAB_CI_BRANCH_TESTS_PYTHON_BASE,
    GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE, GUNICORN_LIMIT_REQUEST_LINE,
    GUNICORN_MAX_REQUESTS, GUNICORN_REQUESTS_JITTER, GUNICORN_TIMEOUT,
    GUNICORN_WORKER_CLASS, GUNICORN_WORKERS
)

from .settings import (
    setting_client_backend_arguments, setting_client_backend_enabled
)
from .utils import load_env_file


class ClientBackend(BaseBackend):
    _loader_module_name = 'client_backends'

    @classmethod
    def get_backend_instance(cls, name):
        backend_class = cls.get(name=name)
        kwargs = setting_client_backend_arguments.value.get(
            name, {}
        )
        return backend_class(**kwargs)

    @classmethod
    def post_load_modules(cls):
        cls.register_url_patterns()
        cls.register_links()
        cls.launch_backends()

    @classmethod
    def launch_backends(cls):
        for backend_name in setting_client_backend_enabled.value:
            cls.get_backend_instance(name=backend_name).launch()

    @classmethod
    def register_links(cls):
        for backend_name in setting_client_backend_enabled.value:
            backend_instance = cls.get_backend_instance(name=backend_name)
            menu_tools.bind_links(
                links=backend_instance.get_links()
            )

    @classmethod
    def register_url_patterns(cls):
        # Hidden import.
        from .urls import urlpatterns

        for backend_name in setting_client_backend_enabled.value:
            backend_instance = cls.get_backend_instance(name=backend_name)

            top_url = '{}/'.format(
                getattr(
                    backend_instance, '_url_namespace',
                    backend_instance.__class__.__name__
                )
            )

            urlpatterns += (
                re_path(
                    route=r'^{}'.format(top_url), view=include(
                        backend_instance.get_url_patterns()
                    )
                ),
            )

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def get_links(self):
        return ()


class Variable:
    def __init__(self, name, default, environment_name):
        self.name = name
        self.default = default
        self.environment_name = environment_name

    def _get_value(self):
        return os.environ.get(self.environment_name, self.default)

    def get_value(self):
        return mark_safe(
            s=self._get_value()
        )


class YAMLVariable(Variable):
    def _get_value(self):
        value = os.environ.get(self.environment_name)
        if value:
            value = yaml_load(stream=value)
        else:
            value = self.default

        return yaml_dump(
            data=value, allow_unicode=True, default_flow_style=True,
            width=999
        ).replace('...\n', '').replace('\n', '')


class PlatformTemplate:
    _registry = {}
    context = {}
    context_defaults = {}
    label = None
    name = None
    settings = None
    template_name = None
    template_string = None
    variables = None

    @classmethod
    def all(cls):
        return cls._registry.values()

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    @classmethod
    def register(cls, klass):
        cls._registry[klass.name] = klass

    def __str__(self):
        return str(
            self.get_label()
        )

    def get_context(self):
        return self.context

    def get_context_defaults(self):
        return self.context_defaults

    def get_label(self):
        return self.label or self.name

    def get_settings_context(self):
        result = {}
        for setting in self.settings or ():
            if setting.value:
                result[setting.global_name] = setting.value

        return result

    def get_template_name(self):
        return self.template_name or 'platform/{}.tmpl'.format(self.name)

    def get_variables_context(self):
        result = {}
        for variable in self.variables or ():
            result[variable.name] = variable.get_value()

        return result

    def render(self, context_string=None):
        """
        context_string allows the management command to pass context to this
        method as a JSON string
        """
        context = {}

        context.update(
            self.get_context_defaults()
        )
        context.update(
            self.get_settings_context()
        )
        context.update(
            self.get_variables_context()
        )
        # get_context goes last to server as the override
        context.update(
            self.get_context()
        )

        if context_string:
            context.update(
                yaml_load(stream=context_string)
            )

        if self.template_string:
            template = Template(template_string=self.template_string)
            return template.render(
                context=Context(dict_=context)
            )
        else:
            return loader.render_to_string(
                context=context, template_name=self.get_template_name()
            )


class PlatformTemplateDockerEntrypoint(PlatformTemplate):
    label = _(message='Template for entrypoint.sh file inside a Docker image.')
    name = 'docker_entrypoint'
    template_name = 'platform/docker/entrypoint.tmpl'

    def get_context(self):
        context = load_env_file()
        context.update(
            {
                'workers': Worker.all()
            }
        )
        return context


class PlatformTemplateDockerComposefile(PlatformTemplate):
    label = _(message='Template that generates the Docker Compose file.')
    name = 'docker_docker_compose'
    template_name = 'platform/docker/docker-compose.yml.tmpl'

    def __init__(self):
        self.variables = (
            Variable(
                name='DEFAULT_DATABASE_NAME',
                default=DEFAULT_DATABASE_NAME,
                environment_name='MAYAN_DEFAULT_DATABASE_NAME'
            ),
            Variable(
                name='DEFAULT_DATABASE_PASSWORD',
                default=DEFAULT_DATABASE_PASSWORD,
                environment_name='MAYAN_DEFAULT_DATABASE_PASSWORD'
            ),
            Variable(
                name='DEFAULT_DATABASE_USER',
                default=DEFAULT_DATABASE_USER,
                environment_name='MAYAN_DEFAULT_DATABASE_USER'
            ),
            Variable(
                name='DEFAULT_ELASTICSEARCH_PASSWORD',
                default=DEFAULT_ELASTICSEARCH_PASSWORD,
                environment_name='MAYAN_DEFAULT_ELASTICSEARCH_PASSWORD'
            ),
            Variable(
                name='DEFAULT_KEYCLOAK_ADMIN',
                default=DEFAULT_KEYCLOAK_ADMIN,
                environment_name='MAYAN_DEFAULT_KEYCLOAK_ADMIN'
            ),
            Variable(
                name='DEFAULT_KEYCLOAK_ADMIN_PASSWORD',
                default=DEFAULT_KEYCLOAK_ADMIN_PASSWORD,
                environment_name='MAYAN_DEFAULT_KEYCLOAK_ADMIN_PASSWORD'
            ),
            Variable(
                name='DEFAULT_KEYCLOAK_DATABASE_HOST',
                default=DEFAULT_KEYCLOAK_DATABASE_HOST,
                environment_name='MAYAN_DEFAULT_KEYCLOAK_DATABASE_HOST'
            ),
            Variable(
                name='DEFAULT_KEYCLOAK_DATABASE_NAME',
                default=DEFAULT_KEYCLOAK_DATABASE_NAME,
                environment_name='MAYAN_DEFAULT_KEYCLOAK_DATABASE_NAME'
            ),
            Variable(
                name='DEFAULT_KEYCLOAK_DATABASE_PASSWORD',
                default=DEFAULT_KEYCLOAK_DATABASE_PASSWORD,
                environment_name='MAYAN_DEFAULT_KEYCLOAK_DATABASE_PASSWORD'
            ),
            Variable(
                name='DEFAULT_KEYCLOAK_DATABASE_USERNAME',
                default=DEFAULT_KEYCLOAK_DATABASE_USERNAME,
                environment_name='MAYAN_DEFAULT_KEYCLOAK_DATABASE_USERNAME'
            ),
            Variable(
                name='DEFAULT_RABBITMQ_PASSWORD',
                default=DEFAULT_RABBITMQ_PASSWORD,
                environment_name='MAYAN_DEFAULT_RABBITMQ_PASSWORD'
            ),
            Variable(
                name='DEFAULT_RABBITMQ_USER',
                default=DEFAULT_RABBITMQ_USER,
                environment_name='MAYAN_DEFAULT_RABBITMQ_USER'
            ),
            Variable(
                name='DEFAULT_RABBITMQ_VHOST',
                default=DEFAULT_RABBITMQ_VHOST,
                environment_name='MAYAN_DEFAULT_RABBITMQ_VHOST'
            ),
            Variable(
                name='DEFAULT_REDIS_PASSWORD',
                default=DEFAULT_REDIS_PASSWORD,
                environment_name='MAYAN_DEFAULT_REDIS_PASSWORD'
            ),
            Variable(
                name='DOCKER_ELASTIC_IMAGE_NAME',
                default=DOCKER_ELASTIC_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_ELASTIC_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_ELASTIC_IMAGE_TAG',
                default=DOCKER_ELASTIC_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_ELASTIC_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_IMAGE_MAYAN_NAME',
                default=DOCKER_IMAGE_MAYAN_NAME,
                environment_name='MAYAN_DOCKER_IMAGE_MAYAN_NAME'
            ),
            Variable(
                name='DOCKER_IMAGE_MAYAN_TAG',
                default=DOCKER_IMAGE_MAYAN_TAG,
                environment_name='MAYAN_DOCKER_IMAGE_MAYAN_TAG'
            ),
            Variable(
                name='DOCKER_KEYCLOAK_IMAGE_NAME',
                default=DOCKER_KEYCLOAK_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_KEYCLOAK_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_KEYCLOAK_IMAGE_TAG',
                default=DOCKER_KEYCLOAK_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_KEYCLOAK_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_KEYCLOAK_POSTGRES_IMAGE_NAME',
                default=DOCKER_KEYCLOAK_POSTGRES_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_KEYCLOAK_POSTGRES_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_KEYCLOAK_POSTGRES_IMAGE_TAG',
                default=DOCKER_KEYCLOAK_POSTGRES_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_KEYCLOAK_POSTGRES_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_IMAGE_NAME',
                default=DOCKER_POSTGRESQL_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_POSTGRESQL_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_IMAGE_TAG',
                default=DOCKER_POSTGRESQL_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_POSTGRESQL_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_RABBITMQ_IMAGE_NAME',
                default=DOCKER_RABBITMQ_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_RABBITMQ_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_RABBITMQ_IMAGE_TAG',
                default=DOCKER_RABBITMQ_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_RABBITMQ_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_REDIS_IMAGE_NAME',
                default=DOCKER_REDIS_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_REDIS_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_REDIS_IMAGE_TAG',
                default=DOCKER_REDIS_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_REDIS_IMAGE_TAG'
            ),
            Variable(
                name='DOCKER_TRAEFIK_IMAGE_NAME',
                default=DOCKER_TRAEFIK_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_TRAEFIK_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_TRAEFIK_IMAGE_TAG',
                default=DOCKER_TRAEFIK_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_TRAEFIK_IMAGE_TAG'
            ),
        )


class PlatformTemplateDockerSupervisord(PlatformTemplate):
    label = _(message='Template for Supervisord inside a Docker image.')
    name = 'docker_supervisord'
    template_name = 'platform/docker/supervisord.tmpl'

    def get_context(self):
        return {
            'OS_USERNAME': DEFAULT_OS_USERNAME,
            'autorestart': 'false',
            'shell_path': '/bin/sh',
            'stderr_logfile': '/dev/fd/2',
            'stderr_logfile_maxbytes': '0',
            'stdout_logfile': '/dev/fd/1',
            'stdout_logfile_maxbytes': '0',
            'workers': Worker.all()
        }


class PlatformTemplateDockerfile(PlatformTemplate):
    label = _(message='Template that generates a Dockerfile file.')
    name = 'docker_dockerfile'
    template_name = 'platform/docker/dockerfile.tmpl'

    def __init__(self):
        self.variables = (
            Variable(
                name='DOCKER_LINUX_IMAGE_VERSION',
                default=DOCKER_LINUX_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_LINUX_IMAGE_VERSION'
            ),
        )


class PlatformTemplateGitLabCI(PlatformTemplate):
    label = _(message='Template that generates a GitLab CI config file.')
    name = 'gitlab-ci'

    def __init__(self):
        self.variables = (
            Variable(
                name='DEFAULT_DATABASE_NAME',
                default=DEFAULT_DATABASE_NAME,
                environment_name='MAYAN_DEFAULT_DATABASE_NAME'
            ),
            Variable(
                name='DEFAULT_DATABASE_PASSWORD',
                default=DEFAULT_DATABASE_PASSWORD,
                environment_name='MAYAN_DEFAULT_DATABASE_PASSWORD'
            ),
            Variable(
                name='DEFAULT_DATABASE_USER',
                default=DEFAULT_DATABASE_USER,
                environment_name='MAYAN_DEFAULT_DATABASE_USER'
            ),
            Variable(
                name='DOCKER_DIND_IMAGE_VERSION',
                default=DOCKER_DIND_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_DIND_IMAGE_VERSION'
            ),
            Variable(
                name='DOCKER_LINUX_IMAGE_VERSION',
                default=DOCKER_LINUX_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_LINUX_IMAGE_VERSION'
            ),
            Variable(
                name='DOCKER_MYSQL_IMAGE_VERSION',
                default=DOCKER_MYSQL_IMAGE_VERSION,
                environment_name='MAYAN_DOCKER_MYSQL_IMAGE_VERSION'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_IMAGE_NAME',
                default=DOCKER_POSTGRESQL_IMAGE_NAME,
                environment_name='MAYAN_DOCKER_POSTGRESQL_IMAGE_NAME'
            ),
            Variable(
                name='DOCKER_POSTGRESQL_IMAGE_TAG',
                default=DOCKER_POSTGRESQL_IMAGE_TAG,
                environment_name='MAYAN_DOCKER_POSTGRESQL_IMAGE_TAG'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_BUILDS_DOCKER',
                default=GITLAB_CI_BRANCH_BUILDS_DOCKER,
                environment_name='MAYAN_GITLAB_CI_BRANCH_BUILDS_DOCKER'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION',
                default=GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION,
                environment_name='MAYAN_GITLAB_CI_BRANCH_BUILDS_DOCUMENTATION'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_BUILDS_PYTHON',
                default=GITLAB_CI_BRANCH_BUILDS_PYTHON,
                environment_name='MAYAN_GITLAB_CI_BRANCH_BUILDS_PYTHON'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO',
                default=GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO,
                environment_name='MAYAN_GITLAB_CI_BRANCH_DEPLOYMENTS_DEMO'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING',
                default=GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING,
                environment_name='MAYAN_GITLAB_CI_BRANCH_DEPLOYMENTS_STAGING'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR',
                default=GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_ALL_MAJOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_ALL_MINOR',
                default=GITLAB_CI_BRANCH_RELEASES_ALL_MINOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_ALL_MINOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR',
                default=GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_DOCKER_MAJOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR',
                default=GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_DOCKER_MINOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION',
                default=GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_DOCUMENTATION'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_NIGHTLY',
                default=GITLAB_CI_BRANCH_RELEASES_NIGHTLY,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_NIGHTLY'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR',
                default=GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_PYTHON_MAJOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR',
                default=GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_PYTHON_MINOR'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_STAGING',
                default=GITLAB_CI_BRANCH_RELEASES_STAGING,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_STAGING'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_RELEASES_TESTING',
                default=GITLAB_CI_BRANCH_RELEASES_TESTING,
                environment_name='MAYAN_GITLAB_CI_BRANCH_RELEASES_TESTING'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_ALL',
                default=GITLAB_CI_BRANCH_TESTS_ALL,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_ALL'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_DOCKER',
                default=GITLAB_CI_BRANCH_TESTS_DOCKER,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_DOCKER'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_PYTHON_ALL',
                default=GITLAB_CI_BRANCH_TESTS_PYTHON_ALL,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_PYTHON_ALL'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_PYTHON_BASE',
                default=GITLAB_CI_BRANCH_TESTS_PYTHON_BASE,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_PYTHON_BASE'
            ),
            Variable(
                name='GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE',
                default=GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE,
                environment_name='MAYAN_GITLAB_CI_BRANCH_TESTS_PYTHON_UPGRADE'
            )
        )


class PlatformTemplateSupervisord(PlatformTemplate):
    label = _(message='Template for Supervisord.')
    name = 'supervisord'

    def __init__(self):
        self.variables = (
            Variable(
                name='GUNICORN_REQUESTS_JITTER',
                default=GUNICORN_REQUESTS_JITTER,
                environment_name='MAYAN_GUNICORN_REQUESTS_JITTER'
            ),
            Variable(
                name='GUNICORN_LIMIT_REQUEST_LINE',
                default=GUNICORN_LIMIT_REQUEST_LINE,
                environment_name='MAYAN_GUNICORN_GUNICORN_LIMIT_REQUEST_LINE'
            ),
            Variable(
                name='GUNICORN_MAX_REQUESTS',
                default=GUNICORN_MAX_REQUESTS,
                environment_name='MAYAN_GUNICORN_MAX_REQUESTS'
            ),
            Variable(
                name='GUNICORN_TIMEOUT',
                default=GUNICORN_TIMEOUT,
                environment_name='MAYAN_GUNICORN_TIMEOUT'
            ),
            Variable(
                name='GUNICORN_WORKER_CLASS',
                default=GUNICORN_WORKER_CLASS,
                environment_name='MAYAN_GUNICORN_WORKER_CLASS'
            ),
            Variable(
                name='GUNICORN_WORKERS',
                default=GUNICORN_WORKERS,
                environment_name='MAYAN_GUNICORN_WORKERS'
            ),
            Variable(
                name='GUNICORN_TIMEOUT',
                default=GUNICORN_TIMEOUT,
                environment_name='MAYAN_GUNICORN_TIMEOUT'
            ),
            Variable(
                name='INSTALLATION_PATH',
                default=DEFAULT_DIRECTORY_INSTALLATION,
                environment_name='MAYAN_INSTALLATION_PATH'
            ),
            Variable(
                name='OS_USERNAME', default=DEFAULT_OS_USERNAME,
                environment_name='MAYAN_OS_USERNAME'
            ),
            Variable(
                name='USER_SETTINGS_FOLDER',
                default=DEFAULT_USER_SETTINGS_FOLDER,
                environment_name='MAYAN_USER_SETTINGS_FOLDER'
            ),
            YAMLVariable(
                name='MEDIA_ROOT', default=settings.MEDIA_ROOT,
                environment_name='MAYAN_MEDIA_ROOT'
            )
        )

    def get_context(self):
        *_, user_settings_folder, media_root = self.variables

        return {
            'autorestart': 'true',
            'shell_path': '/bin/sh',
            'user_settings_folder': Path(
                media_root.get_value()
            ) / user_settings_folder.get_value(),
            'workers': Worker.all()
        }


class PlatformTemplateWorkerQueues(PlatformTemplate):
    label = _(message='Template showing the queues of a worker.')
    name = 'worker_queues'

    variables = (
        Variable(
            name='WORKER_NAME', default=None,
            environment_name='MAYAN_WORKER_NAME'
        ),
    )

    def get_context(self):
        worker_name = self.get_variables_context().get('WORKER_NAME')
        try:
            queues = Worker.get(name=worker_name).queues
        except KeyError:
            raise KeyError(
                'Worker name "{}" not found.'.format(worker_name)
            )

        return {
            'queues': queues, 'queue_names': sorted(
                map(lambda x: x.name, queues)
            )
        }


PlatformTemplate.register(klass=PlatformTemplateDockerEntrypoint)
PlatformTemplate.register(klass=PlatformTemplateDockerComposefile)
PlatformTemplate.register(klass=PlatformTemplateDockerSupervisord)
PlatformTemplate.register(klass=PlatformTemplateDockerfile)
PlatformTemplate.register(klass=PlatformTemplateGitLabCI)
PlatformTemplate.register(klass=PlatformTemplateSupervisord)
PlatformTemplate.register(klass=PlatformTemplateWorkerQueues)



from django.apps import apps
from django.core import checks
from django.test.runner import DiscoverRunner

from mayan.apps.common.apps import MayanAppConfig


def check_app_tests(app_configs=None, **kwargs):
    error_list = []

    runner = DiscoverRunner()

    if app_configs is None:
        app_configs = apps.get_app_configs()

    for app_config in app_configs:
        if issubclass(app_config.__class__, MayanAppConfig):
            has_tests = getattr(app_config, 'has_tests', False)

            test_suite = runner.build_suite(
                test_labels=(app_config.module.__name__,)
            )
            test_suite_test_count = test_suite.countTestCases()

            if has_tests and not test_suite_test_count:
                error_list.append(
                    checks.Warning(
                        hint='Ensure app really defines tests, that the '
                        '`has_tests` attribute is correct, and/or that the '
                        'tests load correctly.',
                        id='mayan_apps.W001',
                        msg='App specifies it has tests but none was found.',
                        obj=app_config
                    )
                )

            if not has_tests and test_suite_test_count:
                error_list.append(
                    checks.Warning(
                        hint='Ensure app really does not define tests, that '
                        'the `has_tests` attribute is correct.',
                        id='mayan_apps.W002',
                        msg='App specifies it does not has tests but some '
                        'were found.',
                        obj=app_config
                    )
                )

    return error_list



from django.conf import settings
from django.contrib import messages
from django.http import Http404
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _, ngettext

from mayan.apps.views.generics import (
    ConfirmView, FormView, MultipleObjectConfirmActionView,
    SingleObjectListView
)

from .forms import SettingForm
from .icons import (
    icon_setting_cluster_configuration_save,
    icon_setting_cluster_namespace_list, icon_setting_namespace_detail,
    icon_setting_edit, icon_setting_revert
)
from .literals import MESSAGE_LOCAL_STORAGE_DISABLED
from .permissions import permission_settings_edit, permission_settings_view
from .settings import setting_cluster


class SettingClusterConfigurationFileSave(ConfirmView):
    post_action_redirect = reverse_lazy(
        viewname='settings:setting_cluster_namespace_list'
    )
    view_icon = icon_setting_cluster_configuration_save
    view_permission = permission_settings_edit

    def dispatch(self, request, *args, **kwargs):
        if settings.COMMON_DISABLE_LOCAL_STORAGE:
            messages.warning(
                message=MESSAGE_LOCAL_STORAGE_DISABLED, request=self.request
            )

        return super().dispatch(request=request, *args, **kwargs)

    def get_extra_context(self):
        return {
            'message': _(
                message='This will overwrite the content of the '
                'configuration file.'
            ),
            'title': _(message='Save settings to the configuration file?')
        }

    def view_action(self, form=None):
        setting_cluster.do_configuration_file_save()

        messages.success(
            message=_(
                message='Settings saved to configuration file successfully.'
            ), request=self.request
        )


class SettingClusterNamespaceListView(SingleObjectListView):
    extra_context = {
        'hide_link': True,
        'title': _(message='Setting namespaces')
    }
    view_icon = icon_setting_cluster_namespace_list
    view_permission = permission_settings_view

    def get_source_queryset(self):
        return setting_cluster.get_namespace_list()


class SettingValueEditView(FormView):
    form_class = SettingForm
    view_icon = icon_setting_edit
    view_permission = permission_settings_edit

    def dispatch(self, request, *args, **kwargs):
        if settings.COMMON_DISABLE_LOCAL_STORAGE:
            messages.warning(
                message=MESSAGE_LOCAL_STORAGE_DISABLED, request=self.request
            )

        self.object = self.get_object()

        return super().dispatch(request=request, *args, **kwargs)

    def form_valid(self, form):
        self.object.do_value_set(
            value=form.cleaned_data['value']
        )
        messages.success(
            message=_(message='Setting updated successfully.'),
            request=self.request
        )
        return super().form_valid(form=form)

    def get_extra_context(self):
        if self.object.get_is_overridden():
            messages.warning(
                message=_(
                    'This setting is overridden by an environment '
                    'variable, changing its value will have no effect.'
                ), request=self.request
            )

        return {
            'hide_link': True,
            'navigation_object_list': ('object', 'setting_namespace'),
            'object': self.object,
            'setting_namespace': self.object.namespace,
            'title': _(message='Edit setting: %s') % self.object
        }

    def get_initial(self):
        return {'setting': self.object}

    def get_object(self):
        return setting_cluster.get_setting(
            global_name=self.kwargs['setting_global_name']
        )

    def get_post_action_redirect(self):
        return reverse(
            viewname='settings:setting_namespace_detail', kwargs={
                'namespace_name': self.object.namespace.name
            }
        )


class SettingValueRevertView(MultipleObjectConfirmActionView):
    pk_url_kwarg = 'setting_global_name'
    success_message_plural = _(
        message='%(count)d settings value reverted.'
    )
    success_message_single = _(
        message='Value of setting "%(object)s" reverted.'
    )
    success_message_singular = _(message='%(count)d setting value reverted.')
    view_icon = icon_setting_revert
    view_permission = permission_settings_edit

    def get_extra_context(self):
        object_list = self.object_list

        result = {
            'message': _(
                message='Unsaved changes will be lost.'
            ),
            'title': ngettext(
                singular='Revert the selected setting value?',
                plural='Revert the selected settings value?',
                number=len(object_list)
            )
        }

        if len(object_list) == 1:
            result['object'] = object_list[0]

        return result

    def get_object_first(self):
        return self.get_object_list()[0]

    def get_object_list(self):
        self.view_mode_single = True
        self.view_mode_multiple = False

        setting = setting_cluster.get_setting(
            global_name=self.kwargs['setting_global_name']
        )

        return (setting,)

    def object_action(self, form, instance):
        instance.do_value_revert()


class SettingNamespaceDetailView(SingleObjectListView):
    view_icon = icon_setting_namespace_detail
    view_permission = permission_settings_view

    def get_extra_context(self):
        namespace = self.get_namespace()

        return {
            'hide_object': True,
            'object': namespace,
            'subtitle': _(
                'Settings inherited from an environment variable take '
                'precedence and cannot be changed in this view. '
            ),
            'title': _(message='Settings in namespace: %s') % namespace
        }

    def get_namespace(self):
        try:
            return setting_cluster.get_namespace(
                name=self.kwargs['namespace_name']
            )
        except KeyError:
            raise Http404(
                _(message='Namespace: %s, not found') % self.kwargs[
                    'namespace_name'
                ]
            )

    def get_source_queryset(self):
        setting_namespace = self.get_namespace()
        return setting_namespace.get_setting_list()



from mayan.apps.dynamic_search.search_models import SearchModel

from .permissions import permission_message_view

search_model_message = SearchModel(
    app_label='messaging', model_name='Message',
    permission=permission_message_view,
    serializer_path='mayan.apps.messaging.serializers.MessageSerializer'
)
search_model_message.add_model_field(field='date_time')
search_model_message.add_model_field(field='subject')
search_model_message.add_model_field(field='body')



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0023_auto_20210821_2059')
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='indexinstancenode',
            unique_together={
                ('index_template_node', 'parent', 'value')
            }
        ),
        migrations.AlterModelOptions(
            name='indextemplate',
            options={
                'ordering': ('label',), 'verbose_name': 'Index template',
                'verbose_name_plural': 'Index templates'
            }
        ),
        migrations.AlterField(
            model_name='indextemplatenode',
            name='expression',
            field=models.TextField(
                help_text="Enter a template to render. Use Django's "
                "default templating language.",
                verbose_name='Indexing expression'
            )
        ),
        migrations.AlterField(
            model_name='indextemplatenode',
            name='index',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='index_template_nodes',
                to='document_indexing.IndexTemplate', verbose_name='Index'
            )
        ),
        migrations.AlterModelOptions(
            name='indextemplatenode',
            options={
                'verbose_name': 'Index template node',
                'verbose_name_plural': 'Index template nodes'
            }
        )
    ]



from io import BytesIO, StringIO
import logging

from django.core.files.base import File
from django.core.files.storage import Storage
from django.utils.deconstruct import deconstructible
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin

from .literals import DEFAULT_STORAGE_BACKEND

logger = logging.getLogger(name=__name__)


class BufferedFile(File):
    def __init__(self, file_object, mode, name=None):
        self.file_object = file_object
        self.mode = mode
        if 'b' in mode:
            self.stream = BytesIO()
        else:
            self.stream = StringIO()

        self.name = name
        self.stream_size = 0

    def close(self):
        self.file_object.close()
        self.stream.close()

    def flush(self):
        return self.file_object.flush()

    def read(self, size=None):
        if size is None:
            size = -1

        if size == -1 or size > self.stream_size:
            while True:
                position = self.stream.tell()
                chunk = self._get_file_object_chunk()
                if chunk:
                    self.stream_size += len(chunk)
                    self.stream.write(chunk)
                    self.stream.seek(position)
                    if self.stream_size >= size and size != -1:
                        break
                else:
                    break

        if size:
            read_size = min(size, self.stream_size)
            self.stream_size -= read_size
        else:
            read_size = None

        return self.stream.read(read_size)


class DefinedStorage(AppsModuleLoaderMixin):
    _loader_module_name = 'storages'
    _registry = {}

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    def __init__(
        self, dotted_path, label, name, kwargs=None, error_message=None
    ):
        self.dotted_path = dotted_path
        self.error_message = error_message
        self.label = label
        self.name = name
        self.kwargs = kwargs or {}
        self.__class__._registry[name] = self

    def __eq__(self, other):
        return True

    def get_storage_instance(self):
        try:
            return self.get_storage_subclass()(**self.kwargs)
        except Exception as exception:
            message = self.error_message or _(
                'Unable to initialize storage: %(name)s. Check the storage '
                'backend dotted path and arguments.'
            ) % {
                'name': self.name
            }

            logger.fatal(message)
            raise TypeError(message) from exception

    def get_storage_subclass(self):
        """
        Import a storage class and return a subclass that will always
        return eq True to avoid creating a new migration when for runtime
        storage class changes.
        """
        try:
            imported_storage_class = import_string(
                dotted_path=self.dotted_path
            )
        except Exception as exception:
            message = self.error_message or _(
                'Unable to initialize storage: %(name)s. Check the storage '
                'backend dotted path and arguments.'
            ) % {
                'name': self.name
            }

            logger.fatal(message)
            raise TypeError(message) from exception

        class DynamicStorageSubclass(imported_storage_class):
            def __init__(self, *args, **kwargs):
                return super().__init__(*args, **kwargs)

            def __eq__(self, other):
                return True

            def deconstruct(self):
                return (
                    'mayan.apps.storage.classes.FakeStorageSubclass', (), {}
                )

        return DynamicStorageSubclass


def defined_storage_proxy_method(method_name):
    def inner_function(self, *args, **kwargs):
        return getattr(
            DefinedStorage.get(name=self.name).get_storage_instance(),
            method_name
        )(*args, **kwargs)

    return inner_function


@deconstructible
class DefinedStorageLazy:
    def __init__(self, name):
        self.name = name
        super().__init__()

    delete = defined_storage_proxy_method(method_name='delete')
    exists = defined_storage_proxy_method(method_name='exists')
    generate_filename = defined_storage_proxy_method(
        method_name='generate_filename'
    )
    open = defined_storage_proxy_method(method_name='open')
    path = defined_storage_proxy_method(method_name='path')
    save = defined_storage_proxy_method(method_name='save')
    size = defined_storage_proxy_method(method_name='size')


class FakeStorageSubclass:
    """
    Placeholder class to allow serializing the real storage subclass to
    support migrations.
    """
    def __eq__(self, other):
        return True


class PassthroughStorage(Storage):
    def __init__(self, *args, **kwargs):
        logger.debug(
            'initializing passthrough storage with: %s, %s', args, kwargs
        )
        next_storage_backend = kwargs.pop(
            'next_storage_backend', DEFAULT_STORAGE_BACKEND
        )
        next_storage_backend_arguments = kwargs.pop(
            'next_storage_backend_arguments', {}
        )

        self.next_storage_class = import_string(
            dotted_path=next_storage_backend
        )

        self.next_storage_backend = self.next_storage_class(
            **next_storage_backend_arguments
        )
        super().__init__(*args, **kwargs)

    def _call_backend_method(self, method_name, kwargs):
        return getattr(self.next_storage_backend, method_name)(**kwargs)

    def delete(self, *args, **kwargs):
        return self.next_storage_backend.delete(*args, **kwargs)

    def exists(self, *args, **kwargs):
        return self.next_storage_backend.exists(*args, **kwargs)

    def path(self, *args, **kwargs):
        return self.next_storage_backend.path(*args, **kwargs)

    def size(self, *args, **kwargs):
        return self.next_storage_backend.size(*args, **kwargs)



from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from .mixins.wizard_mixins import SourceDocumentUploadWizardTestMixin


class SourceDocumentUploadWizardTestCase(
    SourceDocumentUploadWizardTestMixin, GenericDocumentViewTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def test_document_upload_wizard_get_view_no_source_no_permission(self):
        self._clear_events()

        response = self._request_document_upload_wizard_get_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_upload_wizard_get_view_with_source_no_permission(self):
        self._test_source_create()

        self._clear_events()

        response = self._request_document_upload_wizard_get_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_upload_wizard_post_view_with_source_with_document_type_permission(self):
        self._test_source_create()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )

        self._clear_events()

        response = self._request_document_upload_wizard_post_view()
        self.assertEqual(response.status_code, 302)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_upload_wizard_post_view_with_source_no_permission(self):
        self._test_source_create()

        self._clear_events()

        response = self._request_document_upload_wizard_post_view()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.contrib.auth import get_user_model
from django.db import models


class UserLocaleProfileManager(models.Manager):
    def get_by_natural_key(self, user_natural_key):
        User = get_user_model()
        try:
            user = User.objects.get_by_natural_key(user_natural_key)
        except User.DoesNotExist:
            raise self.model.DoesNotExist

        return self.get(user__pk=user.pk)



from django.utils.translation import gettext_lazy as _

from .models import UserMailer
from .permissions import permission_mailing_profile_use


class ObjectEmailActionMixin:
    form_field_widgets = {
        'body': {
            'class': 'django.forms.widgets.Textarea', 'kwargs': {}
        }
    }
    form_fields = {
        'recipient': {
            'label': _(message='Recipient'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    message='Email address of the recipient. Can be '
                    'multiple addresses separated by comma or semicolon. '
                    'A template can be used to reference properties of '
                    'the document.'
                ),
                'required': True
            }
        },
        'cc': {
            'label': _(message='CC'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    message='Address used in the "Bcc" header when '
                    'sending the email. Can be multiple addresses '
                    'separated by comma or semicolon. A template can be '
                    'used to reference properties of the document.'
                ),
                'required': False
            }
        },
        'bcc': {
            'label': _(message='BCC'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    message='Address used in the "Bcc" header when '
                    'sending the email. Can be multiple addresses '
                    'separated by comma or semicolon. A template can be '
                    'used to reference properties of the document.'
                ),
                'required': False
            }
        },
        'reply_to': {
            'label': _(message='Reply to'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    message='Address used in the "Reply-To" header '
                    'when sending the email. Can be multiple addresses '
                    'separated by comma or semicolon. A template can '
                    'be used to reference properties of the document.'
                ),
                'required': False
            }
        },
        'subject': {
            'label': _(message='Subject'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    message='Subject of the email. Can be a string '
                    'or a template.'
                ),
                'required': False
            }
        },
        'body': {
            'label': _(message='Body'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    message='Body of the email to send. Can be a string or '
                    'a template.'
                ),
                'required': False
            }
        },
        'attachment': {
            'label': _(message='Attachment'),
            'class': 'django.forms.BooleanField', 'default': False,
            'help_text': _(
                message='Attach an object to the email.'
            ),
            'required': False
        }
    }
    label = _(message='Send object via email')
    permission = permission_mailing_profile_use

    @classmethod
    def get_form_fields(cls, *args, **kwargs):
        fields = super().get_form_fields(*args, **kwargs)

        fields.update(
            {
                'mailing_profile': {
                    'class': 'mayan.apps.views.fields.FormFieldFilteredModelChoice',
                    'help_text': _(
                        message='Mailing profile to use when sending '
                        'the email.'
                    ),
                    'kwargs': {
                        'source_queryset': UserMailer.objects.filter(
                            enabled=True
                        ), 'permission': cls.permission
                    },
                    'label': _(message='Mailing profile'),
                    'required': True
                }
            }
        )

        return fields

    @classmethod
    def get_form_fieldsets(cls):
        fieldsets = super().get_form_fieldsets()

        fieldsets += (
            (
                _(message='Mailing profile'), {
                    'fields': ('mailing_profile',)
                }
            ), (
                _(message='Parties'), {
                    'fields': ('recipient', 'cc', 'bcc', 'reply_to')
                },
            ), (
                _(message='Content'), {
                    'fields': ('subject', 'attachment', 'body')
                },
            )
        )
        return fieldsets

    def execute(self, context):
        mailing_profile = self.get_mailing_profile()
        mailing_profile.send_object(
            **self.get_execute_data(context=context)
        )

    def get_execute_data(self, context):
        recipient = self.render_field(
            field_name='recipient', context=context
        )
        cc = self.render_field(field_name='cc', context=context)
        bcc = self.render_field(field_name='bcc', context=context)
        reply_to = self.render_field(
            field_name='reply_to', context=context
        )
        subject = self.render_field(
            field_name='subject', context=context
        )
        body = self.render_field(field_name='body', context=context)
        obj = self.get_object(context=context)

        kwargs = {
            'bcc': bcc, 'body': body, 'cc': cc, 'obj': obj,
            'reply_to': reply_to, 'subject': subject, 'to': recipient
        }

        return kwargs

    def get_mailing_profile(self):
        return UserMailer.objects.get(
            pk=self.kwargs['mailing_profile']
        )

    def get_object(self, context):
        return NotImplementedError



from mayan.apps.common.classes import PropertyHelper


class DocumentFileMetadataHelper(PropertyHelper):
    @staticmethod
    @property
    def constructor(*args, **kwargs):
        return DocumentFileMetadataHelper(*args, **kwargs)

    def get_result(self, name):
        file_latest = self.instance.file_latest

        if file_latest:
            return file_latest.get_file_metadata(dotted_name=name)
        else:
            return None


class DocumentFileFileMetadataHelper(PropertyHelper):
    @staticmethod
    @property
    def constructor(*args, **kwargs):
        return DocumentFileFileMetadataHelper(*args, **kwargs)

    def get_result(self, name):
        result = self.instance.get_file_metadata(dotted_name=name)
        return result



from unittest import skip

from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.search import search_model_document
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.dynamic_search.tests.mixins.base import SearchTestMixin

from ..models import Cabinet
from ..permissions import permission_cabinet_view
from ..search import search_model_cabinet

from .mixins import CabinetTestMixin


@skip(reason='Slow test needs to be enabled manually.')
class CabinetSearchFieldSizeLimitTestCase(
    CabinetTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    _test_search_model = search_model_cabinet
    auto_create_test_cabinet = True
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._test_document_stubs = []
        for i in range(1000):
            self._test_document_stubs.append(
                Document(
                    document_type=self._test_document_type,
                    label='test_document_stub_{}'.format(i)
                )
            )
        Document.objects.bulk_create(self._test_document_stubs)

    def test_search_model_cabinet_with_access(self):
        self.grant_access(
            obj=self._test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        ThroughModel = Cabinet.documents.through

        cabinet_documents = []
        for document in Document.objects.all().only('pk'):
            cabinet_documents.append(
                ThroughModel(
                    cabinet_id=self._test_cabinet.pk, document_id=document.pk
                )
            )

        ThroughModel.objects.bulk_create(cabinet_documents)

        self._test_search_backend.index_instance(instance=self._test_cabinet)

        queryset = self._do_test_search(
            query={
                'documents__uuid': str(self._test_document_stubs[0].uuid)
            }
        )
        self.assertTrue(self._test_cabinet in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class CabinetSearchTestCase(
    CabinetTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    _test_cabinet_add_test_document = True
    _test_search_model = search_model_cabinet
    auto_create_test_cabinet = True

    def test_search_model_cabinet_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'documents__label': self._test_document.label
            }
        )
        self.assertTrue(self._test_cabinet not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_cabinet_with_access(self):
        self.grant_access(
            obj=self._test_cabinet, permission=permission_cabinet_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'documents__label': self._test_document.label
            }
        )
        self.assertTrue(self._test_cabinet in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_search_model_document_model_cabinet_with_access(self):
        self.grant_access(
            obj=self._test_cabinet, permission=permission_cabinet_view
        )

        self._test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'documents__label': self._test_document.label
            }
        )
        self.assertTrue(self._test_cabinet in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentCabinetSearchTestCase(
    CabinetTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    _test_cabinet_add_test_document = True
    _test_search_model = search_model_document
    auto_create_test_cabinet = True

    def test_search_model_document_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'cabinets__label': self._test_cabinet.label
            }
        )
        self.assertTrue(self._test_document not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_document_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'cabinets__label': self._test_cabinet.label
            }
        )
        self.assertTrue(self._test_document in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_search_model_document_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'cabinets__label': self._test_cabinet.label
            }
        )
        self.assertTrue(self._test_document not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django import forms
from django.utils.translation import gettext_lazy as _

import mayan
from mayan.apps.acls.models import AccessControlList
from mayan.apps.backends.forms import FormDynamicModelBackend

from .classes import MailerBackend
from .models import UserMailer
from .permissions import permission_mailing_profile_use
from .settings import (
    setting_attachment_body_template, setting_attachment_subject_template,
    setting_document_link_body_template,
    setting_document_link_subject_template
)
from .validators import validate_email_multiple


class ObjectMailForm(forms.Form):
    def __init__(self, *args, **kwargs):
        as_attachment = kwargs.pop('as_attachment', False)
        user = kwargs.pop('user', None)
        super().__init__(*args, **kwargs)

        if as_attachment:
            self.fields[
                'subject'
            ].initial = setting_attachment_subject_template.value

            self.fields[
                'body'
            ].initial = setting_attachment_body_template.value % {
                'project_title': mayan.__title__,
                'project_website': mayan.__website__
            }
        else:
            self.fields[
                'subject'
            ].initial = setting_document_link_subject_template.value
            self.fields['body'].initial = setting_document_link_body_template.value % {
                'project_title': mayan.__title__,
                'project_website': mayan.__website__
            }

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission_mailing_profile_use,
            queryset=UserMailer.objects.filter(enabled=True), user=user
        )

        self.fields['mailing_profile'].queryset = queryset
        try:
            self.fields['mailing_profile'].initial = queryset.get(
                default=True
            )
        except UserMailer.DoesNotExist:
            pass

    email = forms.CharField(
        help_text=_(
            message='Email address of the recipient. Can be multiple '
            'addresses separated by comma or semicolon.'
        ), label=_(message='Email address'), validators=[
            validate_email_multiple
        ]
    )
    subject = forms.CharField(
        label=_(message='Subject'), required=False
    )
    body = forms.CharField(
        label=_(message='Body'), widget=forms.widgets.Textarea(),
        required=False
    )
    mailing_profile = forms.ModelChoiceField(
        help_text=_(
            message='The email profile that will be used to send this email.'
        ), label=_(message='Mailing profile'),
        queryset=UserMailer.objects.none()
    )


class UserMailerBackendSelectionForm(forms.Form):
    backend = forms.ChoiceField(
        choices=(), help_text=_(
            message='The driver to use when sending emails.'
        ), label=_(message='Backend')
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['backend'].choices = MailerBackend.get_choices()


class UserMailerSetupDynamicForm(FormDynamicModelBackend):
    class Meta:
        fields = ('label', 'enabled', 'default')
        model = UserMailer


class UserMailerTestForm(forms.Form):
    email = forms.CharField(
        help_text=_(
            message='Email address of the recipient. Can be multiple '
            'addresses separated by comma or semicolon.'
        ), label=_(message='Email address'), validators=[
            validate_email_multiple
        ]
    )



class SourceException(Exception):
    """Base sources warning"""


class SourceActionException(SourceException):
    """Raised when a source does not have the request action."""


class SourceActionExceptionUnknown(SourceActionException):
    """Raised when a source does not have the request action."""


class SourceActionExceptionInterface(SourceActionException):
    """Base exception for all things related to the action interfaces."""


class SourceActionExceptionInterfaceArgumentMissing(SourceActionExceptionInterface):
    """
    Raised when an action does not supply a required interface argument
    which does not have a default value.
    """


class SourceActionExceptionInterfaceUnknown(SourceActionExceptionInterface):
    """Raised when an action does not support the specified interface."""



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('mailer', '0001_initial')
    ]

    operations = [
        migrations.CreateModel(
            name='UserMailer',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'label', models.CharField(
                        max_length=32, unique=True, verbose_name='Label'
                    )
                ),
                (
                    'default', models.BooleanField(
                        default=True, verbose_name='Default'
                    )
                ),
                (
                    'backend_path', models.CharField(
                        help_text='The dotted Python path to the backend '
                        'class.', max_length=128, verbose_name='Backend path'
                    )
                ),
                (
                    'backend_data', models.TextField(
                        blank=True, verbose_name='Backend data'
                    )
                )
            ],
            options={
                'ordering': ('label',),
                'verbose_name': 'User mailer',
                'verbose_name_plural': 'User mailers'
            }
        ),
        migrations.CreateModel(
            name='UserMailerLogEntry',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True, serialize=False,
                        verbose_name='ID'
                    )
                ),
                (
                    'datetime', models.DateTimeField(
                        auto_now_add=True, verbose_name='Date time'
                    )
                ),
                (
                    'message', models.TextField(
                        blank=True, editable=False, verbose_name='Message'
                    )
                ),
                (
                    'user_mailer', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to='mailer.UserMailer', verbose_name='User mailer'
                    )
                )
            ],
            options={
                'ordering': ('-datetime',),
                'get_latest_by': 'datetime',
                'verbose_name': 'User mailer log entry',
                'verbose_name_plural': 'User mailer log entries'
            }
        )
    ]



import logging

from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.templating.classes import Template

logger = logging.getLogger(name=__name__)


class ExtraDataModelMixin:
    def __init__(self, *args, **kwargs):
        _instance_extra_data = kwargs.pop(
            '_instance_extra_data', {}
        )
        result = super().__init__(*args, **kwargs)
        for key, value in _instance_extra_data.items():
            setattr(self, key, value)

        return result


class ModelMixinConditionField(models.Model):
    condition = models.TextField(
        blank=True, help_text=_(
            'The condition that will determine if this object '
            'is executed or not. Conditions that do not return any value, '
            'that return the Python logical None, or an empty string (\'\') '
            'are considered to be logical false, any other value is '
            'considered to be the logical true.'
        ), verbose_name=_(message='Condition')
    )

    class Meta:
        abstract = True

    def evaluate_condition(self, context):
        if self.has_condition():
            result = Template(template_string=self.condition).render(
                context=context
            ).strip()
            return result
        else:
            return True

    def has_condition(self):
        if self.condition.strip():
            return True
        else:
            return False
    has_condition.help_text = _(
        'The object will be executed, depending on the condition '
        'return value.'
    )
    has_condition.short_description = _(message='Has a condition?')


class ValueChangeModelMixin:
    @classmethod
    def from_db(cls, db, field_names, values):
        new = super().from_db(db=db, field_names=field_names, values=values)
        new._values_previous = dict(
            zip(field_names, values)
        )
        return new

    def __init__(self, *args, **kwargs):
        self._values_previous = kwargs
        super().__init__(*args, **kwargs)

    def _get_field_previous_value(self, field):
        return self._values_previous[field]

    def _has_field_changed(self, field):
        if self._state.adding:
            if getattr(self, field) != self._values_previous[field]:
                return True

        return False



import hashlib
import json
import logging
import os
import threading
import time
import uuid

from django.conf import settings
from django.core.files import locks
from django.utils.encoding import force_bytes

from mayan.apps.storage.settings import setting_temporary_directory

from ..exceptions import LockError

from .base import LockingBackend

lock = threading.Lock()
logger = logging.getLogger(name=__name__)


class FileLock(LockingBackend):
    @classmethod
    def _acquire_lock(cls, name, timeout):
        instance = FileLock(name=name, timeout=timeout)
        return instance

    @classmethod
    def _initialize(cls):
        cls.lock_file = os.path.join(
            setting_temporary_directory.value, hashlib.sha256(
                string=force_bytes(s=settings.SECRET_KEY)
            ).hexdigest()
        )
        open(file=cls.lock_file, mode='a').close()
        logger.debug('lock_file: %s', cls.lock_file)

    @classmethod
    def _purge_locks(cls):
        lock.acquire()
        with open(file=cls.lock_file, mode='r+') as file_object:
            locks.lock(f=file_object, flags=locks.LOCK_EX)
            file_object.seek(0)
            file_object.truncate()
            lock.release()

    def _get_lock_dictionary(self):
        if self.timeout:
            result = {
                'expiration': time.time() + self.timeout,
                'uuid': self.uuid
            }
        else:
            result = {
                'expiration': 0,
                'uuid': self.uuid
            }

        return result

    def _init(self, name, timeout):
        self.name = name
        self.timeout = timeout
        self.uuid = str(
            uuid.uuid4()
        )

        lock.acquire()
        with open(file=self.__class__.lock_file, mode='r+') as file_object:
            locks.lock(f=file_object, flags=locks.LOCK_EX)

            data = file_object.read()

            if data:
                file_locks = json.loads(s=data)
            else:
                file_locks = {}

            if name in file_locks:
                # Someone already got this lock, check to see if it is expired.
                if file_locks[name]['expiration'] and time.time() > file_locks[name]['expiration']:
                    # It expires and has expired, we re-acquired it.
                    file_locks[name] = self._get_lock_dictionary()
                else:
                    lock.release()
                    raise LockError
            else:
                file_locks[name] = self._get_lock_dictionary()

            file_object.seek(0)
            file_object.truncate()
            file_object.write(
                json.dumps(obj=file_locks)
            )
            lock.release()

    def _release(self):
        lock.acquire()
        with open(file=self.__class__.lock_file, mode='r+') as file_object:
            locks.lock(f=file_object, flags=locks.LOCK_EX)
            try:
                file_locks = json.loads(
                    s=file_object.read()
                )
            except EOFError:
                file_locks = {}

            if self.name in file_locks:
                if file_locks[self.name]['uuid'] == self.uuid:
                    file_locks.pop(self.name)
                else:
                    # Lock expired and someone else acquired it
                    pass
            else:
                # Lock expired and someone else released it
                pass

            file_object.seek(0)
            file_object.truncate()
            file_object.write(
                json.dumps(obj=file_locks)
            )
            lock.release()



from django.utils.translation import gettext_lazy as _

from rest_framework.reverse import reverse
from rest_framework_recursive.fields import RecursiveField

from mayan.apps.documents.models.document_models import Document
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from .models import Cabinet
from .permissions import (
    permission_cabinet_add_document, permission_cabinet_remove_document
)


class CabinetSerializer(serializers.ModelSerializer):
    children = RecursiveField(
        help_text=_(message='List of children cabinets.'), label=_(message='Children'),
        many=True, read_only=True
    )
    documents_url = serializers.HyperlinkedIdentityField(
        help_text=_(
            'URL of the API endpoint showing the list documents inside this '
            'cabinet.'
        ), label=_(message='Documents URL'), lookup_url_kwarg='cabinet_id',
        view_name='rest_api:cabinet-document-list'
    )
    full_path = serializers.SerializerMethodField(
        help_text=_(
            'The name of this cabinet level appended to the names of its '
            'ancestors.'
        ), label=_(message='Full path'), read_only=True
    )
    parent_url = serializers.SerializerMethodField(
        label=_(message='Parents URL'), read_only=True
    )

    # This is here because parent is optional in the model but the serializer
    # sets it as required.
    parent = serializers.PrimaryKeyRelatedField(
        allow_null=True, label=_(message='Parent'), queryset=Cabinet.objects.all(),
        required=False
    )

    # DEPRECATION: Version 5.0, remove 'parent' fields from GET request as
    # it is replaced by 'parent_id'.

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _(message='URL'),
                'lookup_url_kwarg': 'cabinet_id',
                'view_name': 'rest_api:cabinet-detail'
            }
        }
        fields = (
            'children', 'documents_url', 'full_path', 'id', 'label',
            'parent_id', 'parent', 'parent_url', 'url'
        )
        model = Cabinet
        read_only_fields = (
            'children', 'documents_url', 'full_path', 'id',
            'parent_id', 'parent_url', 'url'
        )

    def get_full_path(self, obj):
        return obj.get_full_path()

    def get_parent_url(self, obj):
        if obj.parent:
            return reverse(
                viewname='rest_api:cabinet-detail',
                kwargs={'cabinet_id': obj.parent.pk},
                format=self.context['format'],
                request=self.context.get('request')
            )
        else:
            return ''


class CabinetDocumentAddSerializer(serializers.Serializer):
    document = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the document to add to the cabinet.'
        ), label=_(message='Document ID'), source_queryset=Document.valid.all(),
        source_permission=permission_cabinet_add_document
    )


class CabinetDocumentRemoveSerializer(serializers.Serializer):
    document = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the document to remove from the cabinet.'
        ), label=_(message='Document ID'), source_queryset=Document.valid.all(),
        source_permission=permission_cabinet_remove_document
    )



from unittest import mock

from django.urls import reverse

from mayan.apps.testing.tests.base import BaseTestCase

from ..models import DownloadFile, SharedUploadedFile
from ..settings import (
    setting_download_file_expiration_interval,
    setting_shared_uploaded_file_expiration_interval
)

from .mixins import DownloadFileTestMixin, SharedUploadedFileTestMixin


class DownloadFileModelTestCase(DownloadFileTestMixin, BaseTestCase):
    def test_download_file_expiration(self):
        setting_download_file_expiration_interval.do_value_raw_set(
            raw_value=60
        )
        self._create_test_download_file()

        self.assertEqual(
            DownloadFile.objects.get_stale_queryset().count(), 0
        )

        setting_download_file_expiration_interval.do_value_raw_set(
            raw_value=0
        )

        self.assertEqual(
            DownloadFile.objects.get_stale_queryset().count(), 1
        )

    def test_method_get_absolute_url(self):
        self._create_test_download_file()

        self.assertEqual(
            self._test_download_file.get_absolute_url(),
            reverse(viewname='storage:download_file_list')
        )

    def test_stale_deletion_on_error(self):
        self._silence_logger(name='mayan.apps.storage.managers')

        test_case_instance = self

        setting_download_file_expiration_interval.do_value_raw_set(
            raw_value=0
        )

        def fake_method_delete(self, *args, **kwargs):
            if self.pk == test_case_instance._test_download_file_list[0].pk:
                raise Exception

            return super(DownloadFile, self).delete(*args, **kwargs)

        self._create_test_download_file()
        self._create_test_download_file()

        self._clear_events()

        test_download_file_count = DownloadFile.objects.count()

        with mock.patch('mayan.apps.storage.models.DownloadFile.delete', fake_method_delete):
            DownloadFile.objects.stale_delete()

        self.assertEqual(
            DownloadFile.objects.count(),
            test_download_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class SharedUploadedFileManagerTestCase(
    SharedUploadedFileTestMixin, BaseTestCase
):
    def test_shared_uploaded_expiration(self):
        setting_shared_uploaded_file_expiration_interval.do_value_raw_set(
            raw_value=60
        )
        self._create_test_shared_uploaded_file()

        self.assertEqual(
            SharedUploadedFile.objects.get_stale_queryset().count(), 0
        )

        setting_shared_uploaded_file_expiration_interval.do_value_raw_set(
            raw_value=0
        )

        self.assertEqual(
            SharedUploadedFile.objects.get_stale_queryset().count(), 1
        )

    def test_stale_deletion_on_error(self):
        self._silence_logger(name='mayan.apps.storage.managers')

        test_case_instance = self

        setting_shared_uploaded_file_expiration_interval.do_value_raw_set(
            raw_value=0
        )

        def fake_method_delete(self, *args, **kwargs):
            if self.pk == test_case_instance._test_shared_uploaded_file_list[0].pk:
                raise Exception

            return super(SharedUploadedFile, self).delete(*args, **kwargs)

        self._create_test_shared_uploaded_file()
        self._create_test_shared_uploaded_file()

        self._clear_events()

        test_shared_uploaded_file_count = SharedUploadedFile.objects.count()

        with mock.patch('mayan.apps.storage.models.SharedUploadedFile.delete', fake_method_delete):
            SharedUploadedFile.objects.stale_delete()

        self.assertEqual(
            SharedUploadedFile.objects.count(),
            test_shared_uploaded_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

import sentry_sdk
from sentry_sdk.integrations.celery import CeleryIntegration
from sentry_sdk.integrations.django import DjangoIntegration
from sentry_sdk.integrations.redis import RedisIntegration

from django.urls import re_path
from django.utils.translation import gettext_lazy as _

import mayan
from mayan.apps.appearance.classes import Icon
from mayan.apps.common.utils import any_to_bool
from mayan.apps.navigation.classes import Link

from .classes import ClientBackend
from .permissions import permission_test_trigger

logger = logging.getLogger(name=__name__)


class ClientBackendSentry(ClientBackend):
    _url_namespace = 'sentry'

    def get_links(self):
        icon_sentry_debug = Icon(
            driver_name='fontawesome', symbol='bug'
        )

        return (
            Link(
                icon=icon_sentry_debug,
                permission=permission_test_trigger,
                text=_(message='Sentry test error'),
                view='platform:sentry_debug',
            ),
        )

    def get_url_patterns(self):
        def view_trigger_error(request):
            1 / 0

        return [
            re_path(
                route=r'^debug/$', name='sentry_debug',
                view=view_trigger_error
            )
        ]

    def launch(self):
        kwargs = self.setup_arguments()

        kwargs['integrations'] = (
            CeleryIntegration(), DjangoIntegration(), RedisIntegration()
        )

        logger.debug('cleaned arguments: %s', kwargs)

        sentry_instance = sentry_sdk.init(**kwargs)

        logger.debug('client options: %s', sentry_instance._client.options)

    def setup_arguments(self):
        logger.debug('raw arguments: %s', self.kwargs)

        # https://docs.sentry.io/platforms/python/configuration/options/
        options = {}

        # Common Options.
        options['dsn'] = self.kwargs['dsn']

        options['debug'] = any_to_bool(
            value=self.kwargs.get('debug', False)
        )

        options['release'] = mayan.__build_string__

        options['environment'] = self.kwargs.get('environment')

        options['sample_rate'] = float(
            self.kwargs.get('sample_rate', 1.0)
        )

        options['max_breadcrumbs'] = int(
            self.kwargs.get('max_breadcrumbs', 100)
        )

        options['attach_stacktrace'] = any_to_bool(
            value=self.kwargs.get('attach_stacktrace', False)
        )

        options['send_default_pii'] = any_to_bool(
            value=self.kwargs.get('send_default_pii', True)
        )

        options['server_name'] = self.kwargs.get('server_name')

        options['with_locals'] = any_to_bool(
            value=self.kwargs.get('with_locals', True)
        )

        # Transport Options.
        options['transport'] = self.kwargs.get('transport')

        options['http_proxy'] = self.kwargs.get('http_proxy')

        options['https_proxy'] = self.kwargs.get('https_proxy')

        options['shutdown_timeout'] = int(
            self.kwargs.get('shutdown_timeout', 2)
        )

        # Tracing Options.
        options['traces_sample_rate'] = float(
            self.kwargs.get('traces_sample_rate', 0.005)
        )

        return options



from django.apps import apps

from .settings import setting_auto_process


def handler_initialize_new_document_type_file_metadata_settings(
    sender, instance, **kwargs
):
    DocumentTypeSettings = apps.get_model(
        app_label='file_metadata', model_name='DocumentTypeSettings'
    )

    if kwargs['created']:
        DocumentTypeSettings.objects.create(
            auto_process=setting_auto_process.value, document_type=instance
        )


def process_document_file_metadata(sender, instance, **kwargs):
    if instance.document.document_type.file_metadata_settings.auto_process:
        instance.submit_for_file_metadata_processing()



from mayan.apps.testing.tests.base import BaseTestCase

from ..http import URL


class URLTestCase(BaseTestCase):
    def test_query_list_to_string(self):
        url = URL(
            query={
                'a': [1, 2]
            }
        )

        self.assertEqual(url.to_string(), '?a=1&a=2')

    def test_query_string_to_string(self):
        url = URL(
            url='http://example.com?a=1', query={'a': 'string'}
        )

        self.assertEqual(
            url.to_string(), 'http://example.com?a=string'
        )

    def test_query_to_string(self):
        url = URL(
            query={'a': 1}
        )

        self.assertEqual(
            url.to_string(), '?a=1'
        )

    def test_query_list_append_to_string(self):
        url = URL(
            query={'a': '1'}
        )
        url.args.appendlist(key='a', value='2')

        self.assertEqual(
            url.to_string(), '?a=1&a=2'
        )

    def test_query_with_question_mark_to_string(self):
        url = URL(
            query={'a': '1?'}
        )

        self.assertEqual(
            url.to_string(), '?a=1%3F'
        )

    def test_querystring_with_list_to_string(self):
        url = URL(query_string='a=1&a=2')

        self.assertEqual(
            url.args.getlist('a'), ['1', '2']
        )

    def test_querystring_with_question_mark_to_string(self):
        url = URL(query_string='a=1?')

        self.assertEqual(
            url.to_string(), '?a=1%3F'
        )

    def test_querystring_with_question_mark_encoded_to_string(self):
        url = URL(query_string='a=1%3F')

        self.assertEqual(
            url.to_string(), '?a=1%3F'
        )

    def test_querystring_to_args(self):
        url = URL(query_string='a=1')

        self.assertEqual(
            url.args['a'], '1'
        )

    def test_querystring_with_question_mark_encoded_to_args(self):
        url = URL(query_string='a=1%3F')

        self.assertEqual(
            url.args['a'], '1?'
        )

    def test_querystring_mixed_to_args(self):
        url = URL(query_string='a=1&a=2&b=1')

        self.assertEqual(
            url.args.getlist('a'), ['1', '2']
        )
        self.assertEqual(
            url.args.getlist('b'), ['1']
        )

    def test_path_and_querystring_to_string(self):
        url = URL(url='http://example.com', query_string='a=1')

        self.assertEqual(
            url.to_string(), 'http://example.com?a=1'
        )

    def test_path_and_query_to_string(self):
        url = URL(
            url='http://example.com', query={'a': 1}
        )

        self.assertEqual(url.to_string(), 'http://example.com?a=1')

    def test_path_get(self):
        url = URL(url='http://example.com/test_path', query_string='a=1')

        self.assertEqual(url.path, '/test_path')

    def test_path_set(self):
        url = URL(url='http://example.com/test_path', query_string='a=1')

        url.path = 'test_path_2'

        self.assertEqual(url.path, '/test_path_2')

    def test_port_set(self):
        url = URL(scheme='http', netloc='127.0.0.1', port='9999')

        self.assertEqual(
            url.to_string(), 'http://127.0.0.1:9999'
        )

    def test_port_replace(self):
        url = URL(port='9999', url='https://127.0.0.1:8888')

        self.assertEqual(
            url.to_string(), 'https://127.0.0.1:9999'
        )

    def test_scheme_set(self):
        url = URL(scheme='http', netloc='127.0.0.1')

        self.assertEqual(
            url.to_string(), 'http://127.0.0.1'
        )

    def test_scheme_replace(self):
        url = URL(scheme='http', url='https://127.0.0.1')

        self.assertEqual(
            url.to_string(), 'http://127.0.0.1'
        )

    def test_path_viewname_exclusion(self):
        with self.assertRaises(expected_exception=RuntimeError):
            URL(path='/view', viewname='app:viewname')

    def test_path_replace(self):
        url = URL(url='http://127.0.0.1:8000/view_a', path='view_b')

        self.assertEqual(
            url.to_string(), 'http://127.0.0.1:8000/view_b'
        )



from mayan.apps.appearance.classes import Icon

# App

icon_permission = Icon(driver_name='fontawesome', symbol='thumbs-up')

# Group

icon_group_role_list = Icon(driver_name='fontawesome', symbol='user-secret')

# Permission

icon_permission_detail = Icon(driver_name='fontawesome', symbol='thumbs-up')

# Role

icon_role_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='user-secret',
    secondary_symbol='plus'
)
icon_role_single_delete = Icon(driver_name='fontawesome', symbol='times')
icon_role_multiple_delete = icon_role_single_delete
icon_role_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_role_group_list = Icon(driver_name='fontawesome', symbol='users')
icon_role_list = Icon(driver_name='fontawesome', symbol='user-secret')
icon_role_permission_list = Icon(
    driver_name='fontawesome', symbol='thumbs-up'
)



from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_asset_created, event_asset_edited
from ..models import Asset
from ..permissions import (
    permission_asset_create, permission_asset_delete, permission_asset_edit,
    permission_asset_view
)

from .mixins import AssetViewTestMixin


class AssetViewTestCase(AssetViewTestMixin, GenericViewTestCase):
    def test_asset_create_view_no_permission(self):
        asset_count = Asset.objects.count()

        self._clear_events()

        response = self._request_test_asset_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Asset.objects.count(), asset_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_create_view_with_permissions(self):
        self.grant_permission(permission=permission_asset_create)

        asset_count = Asset.objects.count()

        self._clear_events()

        response = self._request_test_asset_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Asset.objects.count(), asset_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_asset)
        self.assertEqual(events[0].verb, event_asset_created.id)

    def test_asset_delete_view_no_permission(self):
        self._create_test_asset()

        asset_count = Asset.objects.count()

        self._clear_events()

        response = self._request_test_asset_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Asset.objects.count(), asset_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_delete_view_with_access(self):
        self._create_test_asset()

        self.grant_access(
            obj=self._test_asset, permission=permission_asset_delete
        )

        asset_count = Asset.objects.count()

        self._clear_events()

        response = self._request_test_asset_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Asset.objects.count(), asset_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_detail_view_no_permission(self):
        self._create_test_asset()

        self._clear_events()

        response = self._request_test_asset_detail_view()
        self.assertNotContains(
            response=response, text=self._test_asset.label, status_code=404
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_detail_view_with_access(self):
        self._create_test_asset()

        self.grant_access(
            obj=self._test_asset, permission=permission_asset_view
        )

        self._clear_events()

        response = self._request_test_asset_detail_view()
        self.assertContains(
            response=response, text=self._test_asset.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_edit_view_no_permission(self):
        self._create_test_asset()

        asset_label = self._test_asset.label

        self._clear_events()

        response = self._request_test_asset_edit_view()
        self.assertEqual(response.status_code, 404)

        self._test_asset.refresh_from_db()
        self.assertEqual(self._test_asset.label, asset_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_edit_view_with_access(self):
        self._create_test_asset()

        self.grant_access(
            obj=self._test_asset, permission=permission_asset_edit
        )

        asset_label = self._test_asset.label

        self._clear_events()

        response = self._request_test_asset_edit_view()
        self.assertEqual(response.status_code, 302)

        self._test_asset.refresh_from_db()
        self.assertNotEqual(self._test_asset.label, asset_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_asset)
        self.assertEqual(events[0].verb, event_asset_edited.id)

    def test_asset_list_view_with_no_permission(self):
        self._create_test_asset()

        self._clear_events()

        response = self._request_test_asset_list_view()
        self.assertNotContains(
            response=response, text=self._test_asset.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_asset_list_view_with_access(self):
        self._create_test_asset()

        self.grant_access(
            obj=self._test_asset, permission=permission_asset_view
        )

        self._clear_events()

        response = self._request_test_asset_list_view()
        self.assertContains(
            response=response, text=self._test_asset.label, status_code=200
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_AJAX_REDIRECTION_CODE, DEFAULT_MAXIMUM_TITLE_LENGTH,
    DEFAULT_MENU_POLLING_INTERVAL, DEFAULT_MESSAGE_POSITION,
    DEFAULT_THROTTLING_MAXIMUM_REQUESTS, DEFAULT_THROTTLING_TIMEOUT
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Appearance'), name='appearance'
)

setting_ajax_redirection_code = setting_namespace.do_setting_add(
    default=DEFAULT_AJAX_REDIRECTION_CODE,
    global_name='APPEARANCE_AJAX_REDIRECTION_CODE', help_text=_(
        'Custom HTTP response code for AJAX redirections.'
    )
)
setting_max_title_length = setting_namespace.do_setting_add(
    default=DEFAULT_MAXIMUM_TITLE_LENGTH,
    global_name='APPEARANCE_MAXIMUM_TITLE_LENGTH', help_text=_(
        'Maximum number of characters that will be displayed as the view '
        'title.'
    )
)
setting_message_position = setting_namespace.do_setting_add(
    choices=(
        'top-left', 'top-center', 'top-right', 'bottom-left',
        'bottom-center', 'bottom-right',
    ), default=DEFAULT_MESSAGE_POSITION,
    global_name='APPEARANCE_MESSAGE_POSITION', help_text=_(
        'Position where the system messages will be displayed.'
    )
)
setting_menu_polling_interval = setting_namespace.do_setting_add(
    default=DEFAULT_MENU_POLLING_INTERVAL,
    global_name='APPEARANCE_MENU_POLLING_INTERVAL', help_text=_(
        'Delay in milliseconds after which the menus will be checked for '
        'updates.'
    )
)
appearance_throttling_maximum_requests = setting_namespace.do_setting_add(
    default=DEFAULT_THROTTLING_MAXIMUM_REQUESTS,
    global_name='APPEARANCE_THROTTLING_MAXIMUM_REQUESTS', help_text=_(
        'Maximum number of requests that can be made before throttling '
        'is enabled.'
    )
)
appearance_throttling_timeout = setting_namespace.do_setting_add(
    default=DEFAULT_THROTTLING_TIMEOUT,
    global_name='APPEARANCE_THROTTLING_TIMEOUT', help_text=_(
        'Time in milliseconds after which a throttled request will '
        'clear allowing an additional request to be performed.'
    )
)



from mayan.apps.rest_api import generics

from ..permissions import (
    permission_workflow_template_edit, permission_workflow_template_view
)
from ..serializers.workflow_template_state_escalation_serializers import WorkflowTemplateStateEscalationSerializer

from .api_view_mixins import ParentObjectWorkflowTemplateStateAPIViewMixin


class APIWorkflowTemplateStateEscalationListView(
    ParentObjectWorkflowTemplateStateAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the workflow template state escalations.
    post: Create a new workflow template state escalation.
    """
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view,
        'POST': permission_workflow_template_edit
    }
    ordering_fields = ('priority', 'id')
    serializer_class = WorkflowTemplateStateEscalationSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'state': self.get_workflow_template_state()
        }

    def get_source_queryset(self):
        return self.get_workflow_template_state().escalations.all()


class APIWorkflowTemplateStateEscalationDetailView(
    ParentObjectWorkflowTemplateStateAPIViewMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Delete the selected workflow template state escalation.
    get: Return the details of the selected workflow template state escalation.
    patch: Edit the selected workflow template state escalation.
    put: Edit the selected workflow template state escalation.
    """
    mayan_object_permission_map = {
        'DELETE': permission_workflow_template_edit,
        'GET': permission_workflow_template_view,
        'PATCH': permission_workflow_template_edit,
        'PUT': permission_workflow_template_edit
    }
    lookup_url_kwarg = 'workflow_template_state_escalation_id'
    serializer_class = WorkflowTemplateStateEscalationSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_source_queryset(self):
        return self.get_workflow_template_state().escalations.all()



from django.db import migrations

from mayan.apps.events.classes import ModelEventType


def code_populate_index_template_event_triggers(apps, schema_editor):
    IndexTemplate = apps.get_model(
        app_label='document_indexing', model_name='IndexTemplate'
    )
    IndexTemplateEventTrigger = apps.get_model(
        app_label='document_indexing', model_name='IndexTemplateEventTrigger'
    )

    for index_template in IndexTemplate.objects.all():
        entries = []

        for key, value in ModelEventType._registry.items():
            # Since migration models are not real, using '.get_for_class'
            # does not work. Instead, directly inspect the model registry.
            if key._meta.model_name == 'document':
                for event_type in value:
                    entries.append(
                        IndexTemplateEventTrigger(
                            index_template=index_template,
                            stored_event_type_id=event_type.get_stored_event_type().pk
                        )
                    )

                IndexTemplateEventTrigger.objects.bulk_create(entries)


def code_populate_index_template_event_triggers_reverse(apps, schema_editor):
    IndexTemplateEventTrigger = apps.get_model(
        app_label='document_indexing', model_name='IndexTemplateEventTrigger'
    )
    IndexTemplateEventTrigger.objects.all().delete()


class Migration(migrations.Migration):
    dependencies = [
        ('document_indexing', '0027_indextemplateeventtrigger')
    ]

    operations = [
        migrations.RunPython(
            code=code_populate_index_template_event_triggers,
            reverse_code=code_populate_index_template_event_triggers_reverse
        )
    ]



class SettingsException(Exception):
    """
    Base exception for the smart_settings app.
    """


class SettingsExceptionRevert(SettingsException):
    """
    Raised when attempting to revert an setting's value.
    """



from django.core.management.base import BaseCommand

from ...search_backends import SearchBackend


class Command(BaseCommand):
    help = 'Upgrade existing search backend persistent structures.'

    def handle(self, *args, **options):
        backend = SearchBackend.get_instance()

        backend.upgrade()



from django.urls import re_path

from .views import (
    SettingClusterConfigurationFileSave, SettingClusterNamespaceListView,
    SettingNamespaceDetailView, SettingValueEditView,
    SettingValueRevertView
)

urlpatterns = [
    re_path(
        route=r'^cluster/configuration/save/$',
        name='setting_cluster_configuration_save',
        view=SettingClusterConfigurationFileSave.as_view()
    ),
    re_path(
        route=r'^cluster/namespaces/$',
        name='setting_cluster_namespace_list',
        view=SettingClusterNamespaceListView.as_view()
    ),
    re_path(
        route=r'^cluster/namespaces/(?P<namespace_name>\w+)/$',
        name='setting_namespace_detail',
        view=SettingNamespaceDetailView.as_view()
    ),
    re_path(
        route=r'^cluster/namespaces/settings/(?P<setting_global_name>\w+)/edit/$',
        name='setting_edit_view', view=SettingValueEditView.as_view()
    ),
    re_path(
        route=r'^cluster/namespaces/settings/(?P<setting_global_name>\w+)/revert/$',
        name='setting_revert_view', view=SettingValueRevertView.as_view()
    )
]



from datetime import datetime

from django.utils.module_loading import import_string
from django.utils.timezone import make_aware

from .settings import setting_gpg_backend, setting_gpg_backend_arguments


class GPGBackend:
    @staticmethod
    def get_instance():
        return import_string(dotted_path=setting_gpg_backend.value)(
            **setting_gpg_backend_arguments.value
        )

    def __init__(self, **kwargs):
        self.kwargs = kwargs


class KeyStub:
    def __init__(self, raw):
        self.fingerprint = raw['keyid']
        self.key_type = raw['type']
        self.date = make_aware(
            value=datetime.fromtimestamp(
                int(
                    raw['date']
                )
            )
        )

        if raw['expires']:
            self.expires = make_aware(
                value=datetime.fromtimestamp(
                    int(
                        raw['expires']
                    )
                )
            )
        else:
            self.expires = None

        self.length = raw['length']
        self.user_id = raw['uids']

    @property
    def key_id(self):
        return self.fingerprint[-8:]


class SignatureVerification:
    def __init__(self, raw):
        self.user_id = raw['username']
        self.status = raw['status']
        self.key_id = raw['key_id']
        self.pubkey_fingerprint = raw['pubkey_fingerprint']

        # Invalid signatures do not have a timestamp attribute.
        if raw['timestamp']:
            self.date_time = make_aware(
                value=datetime.fromtimestamp(
                    int(
                        raw['timestamp']
                    )
                )
            )

        if raw['expire_timestamp']:
            self.expires = make_aware(
                value=datetime.fromtimestamp(
                    int(
                        raw['expire_timestamp']
                    )
                )
            )
        else:
            self.expires = None

        self.fingerprint = raw['fingerprint']
        self.signature_id = raw['signature_id']
        self.stderr = raw['stderr']
        self.trust_level = raw['trust_level']
        self.trust_text = raw['trust_text']
        self.valid = raw['valid']



from django.utils.translation import gettext_lazy as _

argument_help_text_document = _(
    message='Document to which a new file will be uploaded to.'
)

argument_help_text_immediate_mode = _(
    message='When enabled, a document stub is created immediately and '
    'returned. The document file is processed asynchronously. When '
    'disabled, the entire process happens asynchronously with no '
    'returned data. Enabling immediate mode, disables compressed file '
    'processing.'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Documents downloads'), name='document_downloads'
)

event_document_file_downloaded = namespace.add_event_type(
    label=_(message='Document file downloaded'), name='document_file_downloaded'
)



import yaml

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.serialization import yaml_load


class SettingForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setting = self.initial['setting']

        choices = self.setting.get_value_choices()

        if choices:
            self.fields['value'] = forms.ChoiceField(
                choices=list(
                    zip(choices, choices)
                ), required=True, widget=forms.widgets.Select(
                    attrs={'class': 'select2'}
                )
            )
        else:
            self.fields['value'] = forms.CharField(
                required=False, widget=forms.widgets.Textarea()
            )

        self.fields['value'].label = _(message='Value')
        self.fields['value'].help_text = self.setting.help_text or _(
            'Enter the new setting value.'
        )
        self.fields['value'].initial = self.setting.get_value_current()

    def clean(self):
        try:
            yaml_load(
                stream=self.cleaned_data['value']
            )
        except yaml.YAMLError:
            raise ValidationError(
                message=_(
                    '"%s" not a valid entry.'
                ) % self.cleaned_data['value']
            )
        else:
            self.setting.do_value_raw_validate(
                raw_value=self.cleaned_data['value']
            )



from rest_framework import status

from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_mailing_profile_created, event_mailing_profile_edited
from ..models import UserMailer
from ..permissions import (
    permission_mailing_profile_create, permission_mailing_profile_delete,
    permission_mailing_profile_edit, permission_mailing_profile_view
)

from .mixins import MailingProfileAPIViewTestMixin


class MailerAPIViewTestCase(MailingProfileAPIViewTestMixin, BaseAPITestCase):
    def test_mailing_profile_create_api_view_no_permission(self):
        mailing_profile_count = UserMailer.objects.count()

        self._clear_events()

        response = self._request_test_mailing_profile_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(UserMailer.objects.count(), mailing_profile_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mailing_profile_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_mailing_profile_create)

        mailing_profile_count = UserMailer.objects.count()

        self._clear_events()

        response = self._request_test_mailing_profile_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            UserMailer.objects.count(), mailing_profile_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_mailing_profile_created.id)

    def test_mailing_profile_delete_api_view_no_permission(self):
        self._create_test_mailing_profile()

        mailing_profile_count = UserMailer.objects.count()

        self._clear_events()

        response = self._request_test_mailing_profile_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(UserMailer.objects.count(), mailing_profile_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mailing_profile_delete_api_view_with_access(self):
        self._create_test_mailing_profile()

        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_delete
        )

        mailing_profile_count = UserMailer.objects.count()

        self._clear_events()

        response = self._request_test_mailing_profile_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            UserMailer.objects.count(), mailing_profile_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mailing_profile_detail_api_view_no_permission(self):
        self._create_test_mailing_profile()

        self._clear_events()

        response = self._request_test_mailing_profile_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mailing_profile_detail_api_view_with_access(self):
        self._create_test_mailing_profile()
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_view
        )

        self._clear_events()

        response = self._request_test_mailing_profile_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['label'], self._test_mailing_profile.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mailing_profile_edit_api_view_via_patch_no_permission(self):
        self._create_test_mailing_profile()

        mailing_profile_label = self._test_mailing_profile.label

        self._clear_events()

        response = self._request_test_mailing_profile_edit_api_view(
            verb='patch'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_mailing_profile.refresh_from_db()
        self.assertEqual(
            self._test_mailing_profile.label, mailing_profile_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mailing_profile_edit_api_view_via_patch_with_access(self):
        self._create_test_mailing_profile()

        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_edit
        )

        mailing_profile_label = self._test_mailing_profile.label

        self._clear_events()

        response = self._request_test_mailing_profile_edit_api_view(
            verb='patch'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_mailing_profile.refresh_from_db()
        self.assertNotEqual(
            self._test_mailing_profile.label, mailing_profile_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_mailing_profile_edited.id)

    def test_mailing_profile_edit_api_view_via_put_no_permission(self):
        self._create_test_mailing_profile()

        mailing_profile_label = self._test_mailing_profile.label

        self._clear_events()

        response = self._request_test_mailing_profile_edit_api_view(
            verb='put'
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_mailing_profile.refresh_from_db()
        self.assertEqual(
            self._test_mailing_profile.label, mailing_profile_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mailing_profile_edit_api_view_via_put_with_access(self):
        self._create_test_mailing_profile()

        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_edit
        )

        mailing_profile_label = self._test_mailing_profile.label

        self._clear_events()

        response = self._request_test_mailing_profile_edit_api_view(
            verb='put'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_mailing_profile.refresh_from_db()
        self.assertNotEqual(
            self._test_mailing_profile.label, mailing_profile_label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_mailing_profile_edited.id)

    def test_mailing_profile_list_api_view_no_permission(self):
        self._create_test_mailing_profile()

        self._clear_events()

        response = self._request_test_mailing_profile_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_mailing_profile_list_api_view_with_access(self):
        self._create_test_mailing_profile()
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_view
        )

        self._clear_events()

        response = self._request_test_mailing_profile_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['label'],
            self._test_mailing_profile.label
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import hashlib

from django.conf import settings
from django.core import serializers
from django.utils.translation import gettext_lazy as _


class WorkflowStateEscalationBusinessLogicMixin:
    def execute(self, context, workflow_instance):
        if self.evaluate_condition(workflow_instance=workflow_instance):
            try:
                self.get_class_instance().execute(context=context)
            except Exception as exception:
                self.error_log.create(
                    text='{}; {}'.format(
                        exception.__class__.__name__, exception
                    )
                )

                if settings.DEBUG or settings.TESTING:
                    raise
            else:
                self.error_log.all().delete()

    def get_comment(self):
        return self.comment or _(message='Workflow escalation.')

    def get_hash(self):
        return hashlib.sha256(
            string=serializers.serialize(
                format='json', queryset=(self,)
            ).encode()
        ).hexdigest()

    def get_time_display(self):
        return '{} {}'.format(self.amount, self.unit)



from mayan.apps.dynamic_search.tests.mixins.base import SearchTestMixin

from ..permissions import permission_document_view

from .base import GenericDocumentViewTestCase
from .mixins.document_mixins import DocumentViewTestMixin


class DocumentPreviewWidgetViewTestCase(
    DocumentViewTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    def test_document_preview_page_carousel_widget_render(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        response = self._request_test_document_preview_view()
        self.assertContains(
            response=response, status_code=200, text='carousel-container'
        )



"""
Sort existing cabinets after adding:
    class MPTTMeta:
        order_insertion_by = ('label',)
"""
from django.db import migrations

from mptt.models import MPTTModel


def code_sort_existing_cabinets(apps, schema_editor):
    Cabinet = apps.get_model(app_label='cabinets', model_name='Cabinet')

    class ProxyCabinetWithTreeManager(MPTTModel, Cabinet):
        class MPTTMeta:
            order_insertion_by = ('label',)

        class Meta:
            proxy = True

    ProxyCabinetWithTreeManager.objects.rebuild()


class Migration(migrations.Migration):
    dependencies = [
        ('cabinets', '0004_cabinetsearchresult'),
    ]

    operations = [
        migrations.RunPython(
            code=code_sort_existing_cabinets,
            reverse_code=migrations.RunPython.noop
        ),
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Documents'), name='documents'
)

# Document

permission_document_change_type = namespace.add_permission(
    label=_(message='Change type of documents'), name='document_change_type'
)
permission_document_create = namespace.add_permission(
    label=_(message='Create documents'), name='document_create'
)
permission_document_edit = namespace.add_permission(
    label=_(message='Edit documents'), name='document_edit'
)
permission_document_properties_edit = namespace.add_permission(
    label=_(message='Edit document properties'), name='document_properties_edit'
)
permission_document_tools = namespace.add_permission(
    label=_(message='Execute document modifying tools'), name='document_tools'
)
permission_document_view = namespace.add_permission(
    label=_(message='View documents'), name='document_view'
)

# Document file

permission_document_file_delete = namespace.add_permission(
    label=_(message='Delete document files'), name='document_file_delete'
)
permission_document_file_edit = namespace.add_permission(
    label=_(message='Edit document files'), name='document_file_edit'
)
permission_document_file_new = namespace.add_permission(
    label=_(message='Create new document files'), name='document_file_new'
)
permission_document_file_print = namespace.add_permission(
    label=_(message='Print document files'), name='document_file_print'
)
permission_document_file_tools = namespace.add_permission(
    label=_(message='Execute document file modifying tools'),
    name='document_file_tools'
)
permission_document_file_view = namespace.add_permission(
    label=_(message='View document files'),
    name='document_file_view'
)

# Document version

permission_document_version_create = namespace.add_permission(
    label=_(message='Create document versions'),
    name='document_version_create'
)
permission_document_version_delete = namespace.add_permission(
    label=_(message='Delete document versions'),
    name='document_version_delete'
)
permission_document_version_edit = namespace.add_permission(
    label=_(message='Edit document versions'),
    name='document_version_edit'
)
permission_document_version_print = namespace.add_permission(
    label=_(message='Print document versions'), name='document_version_print'
)
permission_document_version_view = namespace.add_permission(
    label=_(message='View document versions'),
    name='document_version_view'
)

# Document type

document_type_namespace = PermissionNamespace(
    label=_(message='Document types'), name='documents_types'
)
permission_document_type_create = document_type_namespace.add_permission(
    label=_(message='Create document types'), name='document_type_create'
)
permission_document_type_delete = document_type_namespace.add_permission(
    label=_(message='Delete document types'), name='document_type_delete'
)
permission_document_type_edit = document_type_namespace.add_permission(
    label=_(message='Edit document types'), name='document_type_edit'
)
permission_document_type_view = document_type_namespace.add_permission(
    label=_(message='View document types'), name='document_type_view'
)

# Trashed document

permission_trashed_document_restore = namespace.add_permission(
    label=_(message='Restore trashed document'), name='document_restore'
)
permission_trashed_document_delete = namespace.add_permission(
    label=_(message='Delete trashed documents'), name='document_delete'
)
permission_document_trash = namespace.add_permission(
    label=_(message='Trash documents'), name='document_trash'
)
permission_trash_empty = namespace.add_permission(
    label=_(message='Empty trash'), name='document_empty_trash'
)



from datetime import datetime
from io import BytesIO
import logging

from django.utils.encoding import force_bytes, force_str
from django.utils.timezone import make_aware

from .classes import GPGBackend
from .exceptions import NeedPassphrase, PassphraseError
from .literals import (
    ERROR_MSG_BAD_PASSPHRASE, ERROR_MSG_GOOD_PASSPHRASE,
    ERROR_MSG_MISSING_PASSPHRASE, KEY_TYPE_SECRET,
    OUTPUT_MESSAGE_CONTAINS_PRIVATE_KEY
)

logger = logging.getLogger(name=__name__)


class KeyBusinessLogicMixin:
    @property
    def key_id(self):
        """
        Short form key ID (using the first 8 characters).
        """
        return self.fingerprint[-8:]

    def introspect_key_data(self):
        # Fix the encoding of the key data stream.
        self.key_data = force_str(s=self.key_data)
        import_results, key_info = GPGBackend.get_instance().import_and_list_keys(
            key_data=self.key_data
        )
        logger.debug('key_info: %s', key_info)

        self.algorithm = key_info['algo']
        self.creation_date = make_aware(
            value=datetime.fromtimestamp(
                int(
                    key_info['date']
                )
            )
        )
        if key_info['expires']:
            self.expiration_date = make_aware(
                value=datetime.fromtimestamp(
                    int(
                        key_info['expires']
                    )
                )
            )
        self.fingerprint = key_info['fingerprint']
        self.length = int(
            key_info['length']
        )
        self.user_id = key_info['uids'][0]
        if OUTPUT_MESSAGE_CONTAINS_PRIVATE_KEY in import_results.results[0]['text']:
            self.key_type = KEY_TYPE_SECRET
        else:
            self.key_type = key_info['type']

    def open(self):
        output_buffer = BytesIO(
            initial_bytes=force_bytes(s=self.key_data)
        )
        return output_buffer

    def sign_file(
        self, file_object, binary=False, clearsign=False, detached=False,
        output=None, passphrase=None
    ):
        """
        Digitally sign a file
        WARNING: using clearsign=True and subsequent decryption corrupts the
        file. Appears to be a problem in python-gnupg or gpg itself.
        https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=55647
        "The problems differ from run to run and file to
        file, and appear to be due to random data being inserted in the
        output data stream."
        """
        file_sign_results = GPGBackend.get_instance().sign_file(
            binary=binary, clearsign=clearsign, detached=detached,
            file_object=file_object, key_data=self.key_data,
            output=output, passphrase=passphrase
        )

        logger.debug(
            'file_sign_results.stderr: %s', file_sign_results.stderr
        )

        if ERROR_MSG_MISSING_PASSPHRASE in file_sign_results.stderr:
            if ERROR_MSG_GOOD_PASSPHRASE not in file_sign_results.stderr:
                raise NeedPassphrase

        if ERROR_MSG_BAD_PASSPHRASE in file_sign_results.stderr:
            raise PassphraseError

        return file_sign_results



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('messaging', '0002_remove_message_parent')
    ]

    operations = [
        migrations.AlterField(
            model_name='message',
            name='sender_content_type',
            field=models.ForeignKey(
                blank=True, null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to='contenttypes.contenttype',
                verbose_name='Sender content type'
            )
        ),
        migrations.AlterField(
            model_name='message',
            name='sender_object_id',
            field=models.PositiveIntegerField(
                blank=True, null=True, verbose_name='Sender object ID'
            )
        )
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('permissions', '0003_remove_role_name')
    ]

    operations = [
        migrations.AlterField(
            model_name='role',
            name='label',
            field=models.CharField(
                help_text='A short text describing the role.',
                max_length=128, unique=True, verbose_name='Label'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Smart links'), name='linking'
)

event_smart_link_created = namespace.add_event_type(
    label=_(message='Smart link created'), name='smart_link_created'
)
event_smart_link_edited = namespace.add_event_type(
    label=_(message='Smart link edited'), name='smart_link_edited'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.converter.layers import layer_decorations
from mayan.apps.converter.links import link_transformation_list

link_decorations_list = link_transformation_list.copy(
    layer=layer_decorations
)
link_decorations_list.text = _(message='Decorations')



from django.core.exceptions import ValidationError as DjangoValidationError

from rest_framework.exceptions import ValidationError
from rest_framework.views import exception_handler


def mayan_exception_handler(exc, context):
    """
    Custom REST API exception handler to ensure all Django ValidationError
    exceptions are converted into HTTP 400 errors and not HTTP 500 errors.
    """
    if isinstance(exc, DjangoValidationError):
        try:
            exc = ValidationError(code=exc.code, detail=exc.message)
        except AttributeError:
            exc = ValidationError(detail=exc.messages)

    return exception_handler(context=context, exc=exc)



from django_celery_beat.models import IntervalSchedule, PeriodicTask

from mayan.apps.common.tests.mixins import ManagementCommandTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..literals import COMMAND_NAME_TASK_MANAGER_PURGE_PERIODIC_TASKS

from .mixins import TaskManagerManagementCommandTestMixin


class TaskManagerManagementCommandTestCase(
    ManagementCommandTestMixin, TaskManagerManagementCommandTestMixin,
    BaseTestCase
):
    _test_management_command_name = COMMAND_NAME_TASK_MANAGER_PURGE_PERIODIC_TASKS

    def test_purge_periodic_tasks_management_command(self):
        self._test_interval_schedule_count = IntervalSchedule.objects.count()
        self._test_periodic_task_count = PeriodicTask.objects.count()

        self._clear_events()

        self._call_test_management_command()

        self.assertEqual(
            IntervalSchedule.objects.count(),
            self._test_interval_schedule_count - 1
        )
        self.assertEqual(
            PeriodicTask.objects.count(),
            self._test_periodic_task_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from shutil import copyfileobj

import sh

from django.utils.translation import gettext_lazy as _

from mayan.apps.dependencies.exceptions import DependenciesException
from mayan.apps.storage.utils import NamedTemporaryFile

from ..classes import MIMETypeBackend

from .literals import DEFAULT_MIMETYPE_PATH


class MIMETypeBackendPerlFileMIMEInfo(MIMETypeBackend):
    def _init(self, copy_length=None, mimetype_path=None):
        self.mimetype_path = mimetype_path or DEFAULT_MIMETYPE_PATH
        self.copy_length = copy_length

        try:
            self.command_mimetype = sh.Command(path=self.mimetype_path).bake(
                '--magic-only'
            )
        except sh.CommandNotFound:
            raise DependenciesException(
                _(message='mimetype not installed or not found.')
            )

    def _get_mime_type(self, file_object, mime_type_only):
        with NamedTemporaryFile() as temporary_file_object:
            file_object.seek(0)
            copyfileobj(
                fsrc=file_object, fdst=temporary_file_object,
                length=self.copy_length
            )
            file_object.seek(0)
            temporary_file_object.seek(0)

            filename, mime_type = self.command_mimetype(
                temporary_file_object.name
            ).split()

            return (mime_type, 'binary')



from furl import furl

from django import forms
from django.apps import apps
from django.contrib import messages
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import classonlymethod
from django.utils.translation import gettext_lazy as _

from formtools.wizard.views import SessionWizardView

from mayan.apps.views.view_mixins import ViewIconMixin

from .classes import DocumentCreateWizardStep
from .exceptions import SourceActionExceptionUnknown
from .icons import (
    icon_document_upload_wizard, icon_wizard_step_first,
    icon_wizard_step_next, icon_wizard_step_previous
)


class DocumentCreateWizard(ViewIconMixin, SessionWizardView):
    template_name = 'appearance/wizard.html'
    view_icon = icon_document_upload_wizard

    @classonlymethod
    def as_view(cls, *args, **kwargs):
        # SessionWizardView needs at least one form in order to be
        # initialized as a view. Declare one empty form and then change the
        # form list in the .dispatch() method.
        class EmptyForm(forms.Form):
            """Empty form"""

        cls.form_list = [EmptyForm]
        return super().as_view(*args, **kwargs)

    def dispatch(self, request, *args, **kwargs):
        Source = apps.get_model(
            app_label='sources', model_name='Source'
        )

        form_list = DocumentCreateWizardStep.get_choices(
            attribute_name='form_class'
        )
        condition_dict = dict(
            DocumentCreateWizardStep.get_choices(attribute_name='condition')
        )

        result = self.__class__.get_initkwargs(
            condition_dict=condition_dict, form_list=form_list
        )
        self.form_list = result['form_list']
        self.condition_dict = result['condition_dict']

        for source in Source.objects.filter(enabled=True):
            try:
                action = source.get_action(name='document_upload')
                if action.has_interface(interface_name='View'):
                    return super().dispatch(request, *args, **kwargs)
            except SourceActionExceptionUnknown:
                """
                Non fatal. Ignore and try the next source.
                """

        messages.error(
            message=_(
                'No interactive document sources have been defined or '
                'none have been enabled, create one before proceeding.'
            ), request=request
        )
        return HttpResponseRedirect(
            redirect_to=reverse(viewname='sources:source_list')
        )

    def get_context_data(self, form, **kwargs):
        context = super().get_context_data(form=form, **kwargs)

        wizard_step = DocumentCreateWizardStep.get(name=self.steps.current)

        context.update(
            {
                'form_css_classes': 'form-hotkey-enter form-hotkey-double-click',
                'step_title': _(
                    'Step %(step)d of %(total_steps)d: %(step_label)s'
                ) % {
                    'step': self.steps.step1,
                    'step_label': wizard_step.label,
                    'total_steps': len(self.form_list)
                },
                'title': _(message='Document upload wizard'),
                'wizard_step': wizard_step,
                'wizard_steps': DocumentCreateWizardStep.get_all()
            }
        )

        context['form_button_overrides'] = (
            {
                'icon': icon_wizard_step_first,
                'label': _(message='First'),
                'name_override': 'wizard_goto_step',
                'value': self.steps.first
            },
            {
                'icon': icon_wizard_step_previous,
                'label': _(message='Previous'),
                'name_override': 'wizard_goto_step',
                'value': self.steps.prev
            },
            {
                'icon': icon_wizard_step_next,
                'is_primary': True,
                'label': _(message='Next')
            }
        )

        if not self.steps.prev:
            context['form_button_overrides'][0]['css_classes'] = 'disabled'
            context['form_button_overrides'][0]['disabled'] = True
            context['form_button_overrides'][1]['css_classes'] = 'disabled'
            context['form_button_overrides'][1]['disabled'] = True

        if not wizard_step.get_next_is_enabled(wizard=self):
            context['form_button_overrides'][2]['css_classes'] = 'disabled'
            context['form_button_overrides'][2]['disabled'] = True

        return context

    def get_form_initial(self, step):
        return DocumentCreateWizardStep.get(
            name=step
        ).get_form_initial(wizard=self) or {}

    def get_form_kwargs(self, step):
        return DocumentCreateWizardStep.get(
            name=step
        ).get_form_kwargs(wizard=self) or {}

    def done(self, form_list, **kwargs):
        query_dict = {}

        for step in DocumentCreateWizardStep.get_all():
            query_dict.update(
                step.done(wizard=self) or {}
            )

        url = furl(
            reverse(viewname='sources:document_upload')
        )
        # Use equal and not .update() to get the same result as using
        # urlencode(doseq=True).
        url.args = query_dict

        return HttpResponseRedirect(
            redirect_to=url.tostr()
        )



from mayan.apps.acls.classes import ModelPermission
from mayan.apps.testing.tests.base import GenericViewTestCase

from ..permissions import permission_events_view

from .mixins import (
    EventObjectTestMixin, EventTypeTestMixin,
    ObjectEventSubscriptionTestMixin, UserEventViewTestMixin,
    UserObjectSubscriptionViewTestMixin
)


class UserEventTypeSubscriptionViewTestCase(
    UserEventViewTestMixin, GenericViewTestCase
):
    def test_user_event_type_subscription_list_view_no_permission(self):
        response = self._request_test_user_event_type_subscription_list_view()
        self.assertEqual(response.status_code, 200)


class UserObjectSubscriptionViewTestCase(
    EventObjectTestMixin, EventTypeTestMixin,
    ObjectEventSubscriptionTestMixin, UserObjectSubscriptionViewTestMixin,
    GenericViewTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_test_object_with_event_type_and_permission()

        self._create_test_object_event_subscription()
        ModelPermission.register(
            model=self.TestModel, permissions=(
                permission_events_view,
            )
        )

    def test_user_event_type_subscription_list_view_no_permission(self):
        response = self._request_user_object_subscription_list_view()
        self.assertNotContains(
            response=response, text=self._test_event_type.label,
            status_code=200
        )
        self.assertNotContains(
            response=response, text=str(self._test_object),
            status_code=200
        )

    def test_user_event_type_subscription_list_view_with_access(self):
        self.grant_access(
            obj=self._test_object, permission=permission_events_view
        )
        response = self._request_user_object_subscription_list_view()
        self.assertContains(
            response=response, text=self._test_event_type.label,
            status_code=200
        )
        self.assertContains(
            response=response, text=str(self._test_object),
            status_code=200
        )



from contextlib import contextmanager
import json

from django.test import tag

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin

from ...models import Source

from ..literals import (
    TEST_CASE_ACTION_NAME_SOURCE_CREATE, TEST_CASE_INTERFACE_NAME_MODEL,
    TEST_SOURCE_BACKEND_PATH_BASE, TEST_SOURCE_LABEL,
    TEST_SOURCE_METADATA_KEY, TEST_SOURCE_METADATA_VALUE
)


@tag('apps_sources')
class SourceTestMixin:
    _test_source_backend_path = None
    _test_source_create_auto = True
    _test_source_file_path = None

    def setUp(self):
        # Initialize the list first. Needed for migration tests.
        self._test_source_list = []

        super().setUp()

        if self._test_source_create_auto:
            self._test_source_create()

    def _execute_test_source_action(self, action_name, extra_data=None):
        backend_data = self.get_test_source_backend_data(
            action_name=action_name, extra_data=extra_data,
            interface_name=TEST_CASE_INTERFACE_NAME_MODEL
        )

        action = self._test_source.get_action(name=action_name)

        return action.execute(
            interface_name='Model', interface_load_kwargs=backend_data
        )

    def _get_test_source_backend_data(self, interface_name, action_name):
        # Method to allow test cases to add their own test source custom
        # data without having to override the main method.
        result = {}

        if action_name == TEST_CASE_ACTION_NAME_SOURCE_CREATE:
            label = self.get_test_source_backend_label()
            result['label'] = label

        return result

    def _test_source_create(self, backend_path=None, extra_data=None):
        self._test_source_pre_create()

        backend_data = self.get_test_source_backend_data(
            action_name=TEST_CASE_ACTION_NAME_SOURCE_CREATE,
            interface_name=TEST_CASE_INTERFACE_NAME_MODEL,
            extra_data=extra_data
        )
        json_backend_data = json.dumps(obj=backend_data)

        backend_path = backend_path or self.get_test_source_backend_path()

        self._test_source = Source.objects.create(
            backend_data=json_backend_data, backend_path=backend_path,
            label=backend_data['label']
        )

        self._test_source_list.append(self._test_source)

    def _test_source_pre_create(self):
        return

    def get_test_source_backend_data(
        self, interface_name, action_name, extra_data=None
    ):
        backend_data = self._get_test_source_backend_data(
            interface_name=interface_name, action_name=action_name
        )

        if extra_data:
            backend_data.update(extra_data)

        return backend_data

    def get_test_source_backend_label(self):
        total_test_source_count = len(self._test_source_list)
        return '{}_{}'.format(TEST_SOURCE_LABEL, total_test_source_count)

    def get_test_source_backend_path(self):
        if self._test_source_backend_path:
            return self._test_source_backend_path
        else:
            return TEST_SOURCE_BACKEND_PATH_BASE

    @contextmanager
    def get_test_source_file_object(self):
        if self._test_source_file_path:
            with open(file=self._test_source_file_path, mode='rb') as file_object:
                yield file_object
        else:
            yield None


class SourceMetadataTestmixin(DocumentTestMixin, SourceTestMixin):
    _test_source_metadata_create_auto = True

    def setUp(self):
        super().setUp()

        if self._test_source_metadata_create_auto:
            self._test_source_metadata_create()

    def _test_source_metadata_create(self):
        self._test_source_metadata = self._test_document_file.source_metadata.create(
            key=TEST_SOURCE_METADATA_KEY, source=self._test_source,
            value=TEST_SOURCE_METADATA_VALUE
        )
